###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4gui

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import sys
import io
import datetime
import functools
import re
import inspect
import webbrowser
import base64

# DS IMPORTS ------------------------------------------------------------------
import pandas as pd
from pandas.api.types import is_string_dtype, is_categorical_dtype
import ipywidgets as widgets
from ipywidgets import Layout
from IPython import get_ipython

from IPython.display import display, Markdown, HTML, Javascript
from traitlets import traitlets

import qgrid
from ipyaggrid import Grid

# DS4N6 IMPORTS ---------------------------------------------------------------

import ds4n6_lib.d4          as d4
import ds4n6_lib.common      as d4com
import ds4n6_lib.gui         as d4gui
import ds4n6_lib.utils       as d4utl

# Tool / Artifact specific libs
import ds4n6_lib.autoruns    as d4_autoruns
import ds4n6_lib.evtx        as d4_evtx
import ds4n6_lib.fstl        as d4_fstl
import ds4n6_lib.kansa       as d4_kansa
import ds4n6_lib.kape        as d4_kape
import ds4n6_lib.macrobber   as d4_macrobber
import ds4n6_lib.mactime     as d4_mactime
import ds4n6_lib.plaso       as d4_plaso
import ds4n6_lib.volatility  as d4_volatility
import ds4n6_lib.tshark      as d4_tshark
import ds4n6_lib.sabonis     as d4_sabonis

###############################################################################
# FUNCTIONS
###############################################################################
curf = ""

def file_select(rootpath="/", readfunc="", compname=None, f2read="", nbsave_prefix="", **kwargs):
    file_folder_select(rootpath, readfunc, compname, f2read, nbsave_prefix, **kwargs)

def folder_select(rootpath="/", readfunc="", compname=None, f2read="", nbsave_prefix="", **kwargs):
    file_folder_select(rootpath, readfunc, compname, f2read, nbsave_prefix, **kwargs)

def file_folder_select(rootpath="/", readfunc="", compname=None, f2read="", nbsave_prefix="", **kwargs):
    """ Function to select a file o folder to read
    """
    # **kwargs are passed through to readfunc()
    def read_selected_file(b, compname=None, f2read=f2read, nbsave_prefix="", **kwargs):
        global curf

        notebook_file = kwargs.get('notebook_file', "")

        if f2read == "":
            with output:
                curf   = d4utl.path
                infile = d4utl.path

                print("Selected file/folder:")
                print("    "+curf)
                print("")
                
                print("- Reading file(s):")
                d4.out = {}
                d4.out = readfunc(infile, **kwargs)
                print("- Reading Done.\n")
                print("- Recover the object read by executing:  myvar = d4.out (e.g. mydf = d4.out)")
                print("")

                savepathtonb = True
                if savepathtonb :
                    print("- Saving f2read path to notebook cell")
                    if notebook_file == "":
                        print("  + notebook_file not defined. Cannot save.")
                    else:
                        hits = d4utl.nbgrep(notebook_file, nbsave_prefix+'_f2read = "')
                        if not hits:
                            cell_a = "# Automatically created - DO NOT EDIT OR REMOVE unless you want to change the file to read (in that case, remove this cell)\n"
                            cell_b = "# "+str(datetime.datetime.now())+"\n"
                            cell_c = '# '+nbsave_prefix+'_f2read = "'+curf+'"'
                            cell   = cell_a + cell_b + cell_c
                            get_ipython().set_next_input(cell)

        else:
            infile = f2read

            print("- Reading file(s):")
            d4.out = {}
            if compname is not None:
                d4.out = readfunc(infile, compname=compname, **kwargs)
            else:
                d4.out = readfunc(infile, **kwargs) 

            print("")

            if d4.out is None:
                print("ERROR: No valid data read. Aborting.")
                return
            else:
                if not len(d4.out) == 0:
                    print("- Reading Done.\n")
                    print("- Recover the object read by executing:  myvar = d4.out (e.g. mydf = d4.out)")
                    print("")


    if f2read != "":
        dummy = False
        read_selected_file(dummy, compname=compname, f2read=f2read, nbsave_prefix=nbsave_prefix, **kwargs)
    else:
        if rootpath == "":
            rootpath = "/"

        display(Markdown("Select "+nbsave_prefix+" file:"))

        if not os.path.exists(rootpath):
            rootpath = "/"
        
        f = d4utl.PathSelector(rootpath)
        display(f.accord)

        button = widgets.Button(description="SELECT & READ")
        output = widgets.Output()
        output.clear_output()
        display(button, output)

        button.on_click(functools.partial(read_selected_file, compname=compname, nbsave_prefix=nbsave_prefix, **kwargs))
        

def xread(**kwargs):
    """ User interface for read data from different type of files
    Syntax: read_data_gui(tool="<tool>" [, rootpath="<rootpath>", notebook_file="<nbfile>", hostname="<hostname>", prefix="<prefix>", ext="<ext>")
    Args:
        tool (str) : tool used to generate files to readd
        rootpath (str) : path of the file/folder with files to read
        notebook_file (str) : path of notebook that is calling the function for check if the data is already readed
        hostname (str) : Hostname of the data to read
        prefix (str) :
        ext (str) :
    """
    def syntax():
        print('Syntax: xread(tool="<tool>" [, rootpath="<rootpath>", notebook_file="<nbfile>", hostname="<hostname>", prefix="<prefix>", ext="<ext>")')
        return

    nargs = len(kwargs)

    if nargs == 0:
        syntax()
        return

    tool          = kwargs.get('tool',          "")
    rootpath      = kwargs.get('rootpath',      "/")
    notebook_file = kwargs.get('notebook_file', "")
    hostname      = kwargs.get('hostname',      "")

    nbsave_prefix = hostname+'_'+tool+'d'

    if notebook_file != "":
        pattern = nbsave_prefix+'_f2read = "/.*$'

        print("- Searching notebook for saved input file / folder ("+nbsave_prefix+")")
        hits = d4utl.nbgrep(notebook_file, pattern)
        if hits:
            f2read = hits[0].split(" = ")[1].strip('"').strip('# ')
            print("  + Found: "+f2read)
        else:
            f2read = ""
            print("  + Not Found")
        print("")
    else:
        f2read = ""
    
    # TODO: possible bug when using relative paths in xread

    func2exec = 'd4_'+tool+'.read_data'

    d4gui.folder_select(rootpath, functools.partial(eval(func2exec), **kwargs), f2read=f2read, nbsave_prefix=nbsave_prefix, notebook_file=notebook_file)
        
        
        

# dfmenu() & friends ##########################################################
def xmenu(obj, engine="default"):
    """ User interface to explore forensics dataframes and dictionaries of dataframes

        Args: 
            obj (pandas.DataFrame or dict of pandas.DataFrame): data to exploer
            engine (str) : type of grid to show the data, options: default|qgrid|aggrid
                (default is "default")
                
        Returns: 
            dfs_explorer: instance that display the UI
    """
    if len(obj) == 0:
        print("- Empty object ({})".format(type(obj)))
        return

    objtype = d4com.data_identify(obj)
    column_mode = False
    if re.search("^pandas_dataframe-", objtype):
        dfs = dict({" + DataFrame": obj})
        for col in obj.columns:
            dfs[" -- "+col] = obj[col].to_frame()
        column_mode = True
    elif re.search("^dict-", objtype) and len(obj) > 0:
        dfs = obj
    else:
        print("- Unsupported object ({})".format(type(obj)))
        return

    explorer = dfs_explorer(dfs, column_mode=column_mode)
    explorer.display()
    return explorer 


# use:
#
#   grid = d4gui.aggrid(df)
#   grid
#
# get selected rows:
#    grid.grid_data_out.get('rows')
def aggrid(df, page_size=10):
    """ Function description

        Args: 
            df (pandas.DataFrame): data to show
            page_size (int): number of rows per page
                (default is 10)

        Returns: 
            ipyaggrid.Grid: object that show the data using ag-grid JS framework
        
    """
    timestamp_cellrenderer='''
    function (params){
        return params.value;
    }
    '''
    column_defs = [{'headerName': df.index.name, 'field': df.index.name, 'checkboxSelection': True, 'cellRenderer': timestamp_cellrenderer if df.index.name == "timestamp" else None}]
    column_defs +=  [{'headerName': c, 'field': c, 'cellRenderer': timestamp_cellrenderer if c == "timestamp" else None} for c in df.columns]
    gridOptions = {
        'columnDefs': column_defs,
        'defaultColDef': {'sortable': True, 'filter': True, 'resizable': True, 'floatingFilter': True},
        'animateRows': True,
        'enableRangeSelection': True,
        'enableColResize': True,
        'rowSelection': 'multiple',
        'enableFilter': True,
        'enableSorting': True,
        'pagination':True,
        'paginationPageSize': page_size
    }
    grid = Grid(
        grid_options=gridOptions, 
        grid_data=df,
        quick_filter=True,
        columns_fit="auto",
        export_csv=True,
        export_excel=True,
        export_mode='auto',
        export_to_df=True,
        keep_multiindex=False,
        theme="ag-theme-balham",
        index=True 
    )
    return grid

class dfs_explorer():
    """ A class used to display a dataframes explorer

    Atributes:
        current_df (pandas.DataFrame) :
        current_df_name (str) :
        simple (bool) : 
    
    Methods: 
        update_options()
        update_grid()
        reload_df(name)
        display()
        export_grid_to_df()
    """
    current_df=None
    current_df_name=""
    simple=True
    

    def __init__(self, dfs, grid_type="default", column_mode=False):
        self.dfs = {}
        for key in dfs.keys():
            if not dfs[key].empty:
                self.dfs[key] = dfs[key]
        # self.dfs = dfs
        if grid_type in ['default', 'qgrid', 'aggrid']: 
            self.grid_type = grid_type
        box_layout = Layout(overflow='auto', width='99%')
        self.dfout = widgets.Output(layout=box_layout)     
        self.dfout.clear_output()
        self.footer = widgets.Output(layout=box_layout)

        def simple_output_selector_eventhandler(change):
            self.simple = change.new
            self.reload_df(self.current_df_name)

        def select_df_eventhandler(change):
            self.reload_df(change.new)
        
        def select_grid_eventhandler(change):
            self.grid_type = change.new
            self.reload_df(self.current_df_name)
        
        def mixs(num): 
            try: 
                ele = int(num) 
                return (0, ele, '') 
            except ValueError: 
                return (1, num, '') 

        dropdown_description = 'Select DataFrame'
        self.column_mode = column_mode
        if column_mode:
            dropdown_description += " or Column"
        options=[dropdown_description]
        dfs_keys = list(self.dfs.keys())
        dfs_keys.sort(key = mixs)
        options.extend(dfs_keys)
        self.dropdown = widgets.Dropdown(layout={'width': '550px'}, options=options, description=dropdown_description, disabled=False, style={'description_width': 'initial'})
        self.dropdown.observe(select_df_eventhandler, names='value')

        grid_options=['aggrid', 'qgrid', 'default']
        self.grid_selection = widgets.Dropdown(layout={'width': '250px'}, options=grid_options, description='Select grid', disabled=False, value=self.grid_type)
        self.grid_selection.observe(select_grid_eventhandler, names='value')

        self.max_rows = 20
        max_rows_options=['20', '50', '100', '200', '500', '1000']
        def select_max_rows_eventhandler(change):
            self.max_rows = int(change.new)
            self.reload_df(self.current_df_name)
        self.max_rows_selection = widgets.Dropdown(layout={'width': '250px'}, options=max_rows_options, description='Select max_rows', disabled=False, value=str(self.max_rows), style={'description_width': 'initial'})
        self.max_rows_selection.observe(select_max_rows_eventhandler, names='value')

        #if re.search("^pandas_dataframe-evtx_file_df", objtype) or re.search("^pandas_dataframe-evtx_file_raw_df", objtype):
        self.simple_check = widgets.Checkbox(self.simple, description='Simple Output')
        self.simple_check.observe(simple_output_selector_eventhandler, names='value')
 
        self.consolidate_cols = True
        self.consolidate_cols_check = widgets.Checkbox(self.consolidate_cols, description='Consolidate cols.')    
        def consolidate_cols_check_eventhandler(change):
            self.consolidate_cols = change.new
            self.reload_df(self.current_df_name)
        self.consolidate_cols_check.observe(consolidate_cols_check_eventhandler, names='value')
        
        self.collapse_constant_cols = True
        self.collapse_constant_cols_check = widgets.Checkbox(self.collapse_constant_cols, description='Collapse constant cols.')
        def collapse_constant_cols_check_eventhandler(change):
            self.collapse_constant_cols = change.new
            self.reload_df(self.current_df_name)
        self.collapse_constant_cols_check.observe(collapse_constant_cols_check_eventhandler, names='value')
        
        self.hide_cols = True
        self.hide_cols_check = widgets.Checkbox(self.hide_cols, description='Hide cols.')
        def hide_cols_check_eventhandler(change):
            self.hide_cols = change.new
            self.reload_df(self.current_df_name)
        self.hide_cols_check.observe(hide_cols_check_eventhandler, names='value')
        
        
        self.apply_filters = True
        self.apply_filters_check = widgets.Checkbox(self.apply_filters, description='Apply Filters')
        def apply_filters_check_eventhandler(change):
            self.apply_filters = change.new
            self.reload_df(self.current_df_name)
        self.apply_filters_check.observe(apply_filters_check_eventhandler, names='value')
        
        self.wrap_cols = True
        self.wrap_cols_check = widgets.Checkbox(self.wrap_cols, description='Wrap cols.')
        def wrap_cols_check_eventhandler(change):
            self.wrap_cols = change.new
            self.reload_df(self.current_df_name)
        self.wrap_cols_check.observe(wrap_cols_check_eventhandler, names='value')
        
        
        self.out = True        
        self.out_df = False        
        
        self.beautify_cols = False
        self.beautify_cols_check = widgets.Checkbox(self.beautify_cols, description='Beautify cols.')
        def beautify_cols_check_eventhandler(change):
            self.beautify_cols = change.new
            self.reload_df(self.current_df_name)
        self.beautify_cols_check.observe(beautify_cols_check_eventhandler, names='value')
        
        self.ret = True        
        self.ret_out = True
        self.ret_out_check = widgets.Checkbox(self.ret_out, description='Ret out.')
        def ret_out_check_eventhandler(change):
            self.ret_out = change.new
            self.reload_df(self.current_df_name)
        self.ret_out_check.observe(ret_out_check_eventhandler, names='value')
        
        self.menu = widgets.Box([
            self.dropdown,
            self.grid_selection,
            self.simple_check,
            self.max_rows_selection
        ])

        self.options_dict = {
            'consolidate_cols': self.consolidate_cols_check,
            'collapse_constant_cols': self.collapse_constant_cols_check,
            'hide_cols': self.hide_cols_check,
            'apply_filters': self.apply_filters_check,    
            'wrap_cols': self.wrap_cols_check,    
            'beautify_cols': self.beautify_cols_check,
            'ret_out': self.ret_out_check,
        }
        
        self.select_df_button = widgets.Button(description="Export Grid to DataFrame", icon="download", layout={'width': '250px'})
        def select_df_button_eventhandler(change):
            self.export_grid_to_df()
        self.select_df_button.on_click(select_df_button_eventhandler)
        
        # Loading DF section
        loading_label = widgets.HTML("<div style='color: orange; font-weight: bold;'>Data Loading. Please Wait…</div>")
        self.loading = widgets.HBox([loading_label])
        self.loading.layout.visibility = 'hidden'
        self.loading.layout.height = '0px'

        # Selected DF Section
        self.selected_df_label = widgets.Label(value="Selected dataframe: ")
        self.selected_df = widgets.HBox([self.selected_df_label])


    def update_options(self):
        if not self.current_df_name:
            return
        source_type = d4com.whatis(self.dfs[self.current_df_name]) 
        if "-" in source_type:
            source_type = source_type.split("-")[1]
        if "_" in source_type:
            source_type = source_type.split("_")[0]
        srcoptions = d4com.get_source_options(source_type)
        options_items = [self.options_dict['collapse_constant_cols'], self.options_dict['hide_cols'], self.options_dict['ret_out']]
        for option in srcoptions:
            options_items.append(self.options_dict[option])
        self.simple_options = widgets.GridBox(options_items, layout=widgets.Layout(grid_template_columns="repeat(5, 200px)"))

    def update_grid(self):
        if not self.current_df_name :
            return
        if self.simple and not self.ret_out:
            self.grid = None
        elif self.grid_type == 'aggrid':
            try:
                for col in self.current_df.columns:
                    if self.current_df[col].dtype == 'Int64':
                        self.current_df[col] = self.current_df[col].astype("int64")
                self.grid = aggrid(self.current_df, page_size=self.max_rows)
            except:
                display(Markdown("<span style=\"color:red\">*DF unavailable to show in aggrid, please use default or qgrid*</span>"))
                return
        elif self.grid_type == 'qgrid':
            if not 'Selected' in self.current_df.columns:
                self.current_df.insert(0,'Selected', False)
            # qgrid filter issue
            for col in self.current_df.columns:
                if is_string_dtype(self.current_df[col]):
                    self.current_df[col] = self.current_df[col].astype("object")
                if is_categorical_dtype(self.current_df[col]) and len(self.current_df[col].cat.categories) == 0:
                    self.current_df[col].cat.add_categories("", inplace=True)
            self.grid = qgrid.show_grid(self.current_df, show_toolbar=True, grid_options={'forceFitColumns':False, 'maxVisibleRows': self.max_rows})
        else:
            pd.set_option("display.max_rows", self.max_rows)
            pd.set_option("display.min_rows", self.max_rows)
            pd.set_option('display.latex.escape', False)
            self.grid = self.current_df
        with self.dfout:
            if not self.simple:
                display(Markdown("  * No. Entries: {}".format(len(self.current_df))))
            if self.grid is not None:
                display(self.grid)
                display(Markdown("  * Access your selected DataFrame via the d4.out variable"))
                display(self.select_df_button)
        

    def reload_df(self, name):
        self.loading.layout.visibility = 'visible'
        self.loading.layout.height = '30px'
        self.selected_df_label.value = "Selected DataFrame: {}".format(name)
        self.dfout.clear_output()
        self.footer.clear_output()
        if str(name).startswith("Select DataFrame") or name == "":
            self.current_df_name = ""
            self.current_df = None
            return
        with self.dfout:
            self.current_df_name=name
            self.update_options()
            if self.column_mode and self.simple and len(self.dfs[name].columns) == 1:
                self.current_df = self.dfs[name]
                self.current_df[self.current_df.columns[0]].simple()
            else:
                display(Markdown("**Simple Options:**"))
                display(self.simple_options)
                if self.simple:            
                    self.current_df=self.dfs[name].spl(
                        consolidate_cols=self.consolidate_cols,
                        collapse_constant_cols=self.collapse_constant_cols,
                        hide_cols=self.hide_cols,
                        apply_filters=self.apply_filters,
                        out=self.out,
                        out_df=self.out_df,
                        beautify_cols=self.beautify_cols,
                        ret=self.ret,
                        ret_out=self.ret_out
                    )
                else:
                    self.current_df = self.dfs[name]             
                self.update_grid()
        self.loading.layout.visibility = 'hidden'
        self.loading.layout.height = '0px'
            
    def display(self):   
        display(Markdown("**DataFrame visualization menu:**"))
        display(self.menu)
        display(self.selected_df)
        display(self.loading)
        display(self.dfout)
        display(self.footer)

    def export_grid_to_df(self):
        self.footer.clear_output()
        if not self.current_df_name :
            with self.footer:
                display(Markdown("_Nothing to export_"))
            return
        if self.grid_type == 'aggrid':
            d4.out = self.grid.grid_data_out.get('rows')
            message = "d4.out variable updated with selected rows from '{}' dataframe".format(self.current_df_name)
            if d4.out is None or d4.out.size == 0:
                d4.out = self.grid.grid_data_out.get("grid")
                message = "d4.out variable updated with '{}' dataframe".format(self.current_df_name)
        elif self.grid_type == 'qgrid':
            df_changed = self.grid.get_changed_df()
            df_selected = df_changed[df_changed['Selected']]
            if df_selected.empty:
                d4.out = df_changed.drop('Selected', axis=1)
                message = "d4.out variable updated with grid content from '{}' dataframe".format(self.current_df_name)
            else:
                d4.out = df_selected.drop('Selected', axis=1)
                message = "d4.out variable updated with selected rows from '{}' dataframe".format(self.current_df_name)
        else:
            d4.out = self.current_df
            message = "d4.out variable updated with '{}' dataframe".format(self.current_df_name)
        with self.footer:
            display(Markdown("_{}_".format(message)))
            display(d4.out.info())


def create_pd_int_option(name):
    field = widgets.IntText(
        value=pd.get_option(name),
        description='{}: '.format(name),
        disabled=False,
        style={'description_width': 'initial'}
    )
    def field_change_eventhandler(change):
        pd.set_option(name, change.new)

    field.observe(field_change_eventhandler, names='value')
    return field

def create_pd_boolean_option(name):
    field = widgets.Checkbox(
        indent=False,
        value=pd.get_option(name),
        description=name,
        disabled=False,
        style={'description_width': 'initial'}
    )
    def field_change_eventhandler(change):
        pd.set_option(name, change.new)

    field.observe(field_change_eventhandler, names='value')
    return field

def create_pd_text_option(name, options):
    field = widgets.Dropdown(
        options=options,
        value=pd.get_option(name),
        description='{}: '.format(name),
        disabled=False,
        style={'description_width': 'initial'}
    )
    def field_change_eventhandler(change):
        pd.set_option(name, change.new)

    field.observe(field_change_eventhandler, names='value')
    return field

def xdisplay():
    """ User interface to change Pandas display options 
    
    Display options that you can change:
    - display.max_rows
    - display.min_rows
    - display.max_columns
    - display.colheader_justify
    - display_expand_frame_repr
    """
    display_max_rows_field = create_pd_int_option('display.max_rows')
    display_min_rows_field = create_pd_int_option('display.min_rows')
    display_max_columns_field = create_pd_int_option('display.max_columns')
    display_colheader_justify_field = create_pd_text_option('display.colheader_justify', ['right', 'left'])
    display_expand_frame_repr_field = create_pd_boolean_option('display.expand_frame_repr')
    display(Markdown('#### Pandas options'))
    display(display_max_rows_field)
    display(display_min_rows_field)
    display(display_max_columns_field)
    display(display_colheader_justify_field)
    display(display_expand_frame_repr_field)

def xanalysis(dfs):
    """ User Interface to show analysis from forensics dataframes

        Args: 
            dfs (dict of pandas.DataFrame): data for analysis
    """
    # get analysis options
    def get_analysis_options(dfs):
        old_stdout = sys.stdout
        new_stdout = io.StringIO()
        sys.stdout = new_stdout
        d4com.anl(dfs)
        anl_output = new_stdout.getvalue()
        sys.stdout = old_stdout
        lines = anl_output.split("\n")
        options = [("Select analysis type", " ")]
        for line in lines:
            if line.startswith("-") and not line.startswith("- No analysis modules"):
                sep = line.find(":")
                value = line[2:sep]
                description = line[sep+1:].split("(")[0].strip()
                options.append((description, value))
        return options
    anl_result = None
    object_selected = None
    anl_src = dfs
    dfs_src = dfs
    box_layout = Layout(overflow='auto', width='99%')
    anl_out = widgets.Output(layout=box_layout) 
    obj_type = d4com.data_identify(dfs)
    object_options = []
    dataframe_options = ['Select one...']
    def df_type(df):
        return d4com.data_identify(df).split("-")[1]

    if obj_type.startswith("dict"):
        object_options.append('Select one...')
        object_options.append('Dictionary')
        dataframe_options += tuple(dfs.keys())
        type_options = ['Select one...']
        type_options += tuple(set([df_type(dfs[df_id]) for df_id in dfs]))
        is_df = False
    else:
        type_options = [df_type(dfs), ]
        is_df = True
    object_options.append('DataFrame')
    

    object_selection = widgets.Dropdown(layout={'width': '250px'}, options=object_options, description='Analysis object: ', disabled=is_df, style={'description_width': 'initial'})
    def object_selection_eventhandler(change):
        nonlocal object_selected
        nonlocal type_selection
        nonlocal dataframe_selection
        nonlocal second_options
        nonlocal is_df
        object_selected=change.new
        if change.new == "Dictionary":
            type_selection.layout.visibility='visible'
            type_selection.value="Select one..."
            dataframe_selection.layout.visibility='hidden'
            second_options.layout.visibility='hidden'   
        elif change.new == "DataFrame":
            type_selection.layout.visibility='visible'
            type_selection.value="Select one..."
            dataframe_selection.layout.visibility='visible'
            dataframe_selection.value="Select one..."
            if not is_df:
                second_options.layout.visibility='hidden'   
        else:
            type_selection.layout.visibility='hidden'
            dataframe_selection.layout.visibility='hidden'
            second_options.layout.visibility='hidden'   
    object_selection.observe(object_selection_eventhandler, names='value')

    type_selection = widgets.Dropdown(layout={'width': '250px'}, options=type_options, description='Analysis type: ', disabled=is_df, style={'description_width': 'initial'})
    if is_df:
        type_selection.layout.visibility='visible'
    else:
        type_selection.layout.visibility='hidden'
    def type_selection_eventhandler(change):
        nonlocal object_selected
        nonlocal dataframe_selection
        nonlocal analysis_selection
        nonlocal second_options
        nonlocal anl_src
        nonlocal anl_out
        if change.new != "Select one...":
            anl_src = d4utl.extract_dfs_from_type("pandas_dataframe-"+change.new, dfs)
            if object_selected == "Dictionary":
                analysis_selection.options = get_analysis_options(anl_src)
                if len(analysis_selection.options) == 1:
                    anl_out.clear_output()
                    with anl_out:
                        display(Markdown("**No analysis available**"))
                else:
                    second_options.layout.visibility='visible'   
                    anl_out.clear_output()
            elif object_selected == "DataFrame":
                dataframe_selection.options = ['Select one...', ]
                dataframe_selection.options += tuple(anl_src.keys())
                   
    type_selection.observe(type_selection_eventhandler, names='value')

    
    dataframe_selection = widgets.Dropdown(layout={'width': '450px'}, options=dataframe_options, description='DF to analyze: ', disabled=is_df, style={'description_width': 'initial'})
    dataframe_selection.layout.visibility='hidden'
    def dataframe_selection_eventhandler(change):
        nonlocal anl_src
        nonlocal dfs_src
        nonlocal second_options
        nonlocal analysis_selection
        nonlocal anl_out
        if change.new != "Select one...":
            anl_src = dfs_src[change.new]
            analysis_selection.options = get_analysis_options(anl_src) 
            if len(analysis_selection.options) == 1:
                anl_out.clear_output()
                with anl_out:
                    display(Markdown("**No analysis available**"))
            else:
                second_options.layout.visibility='visible'  
                anl_out.clear_output()
        else: 
            second_options.layout.visibility='hidden'  
    dataframe_selection.observe(dataframe_selection_eventhandler, names='value')

    analysis_selection = widgets.Dropdown(layout={'width': '550px'}, options=[], description='Available analysis types: ', disabled=False, style={'description_width': 'initial'})
    def select_analysis_eventhandler(change):
        nonlocal anl_result 
        nonlocal anl_src
        nonlocal anl_out
        anl_out.clear_output()
        anl_result = None
        if change.new !=" ":
            with anl_out:
                anl_result = d4com.anl(anl_src,change.new)
                display(anl_result)

    analysis_selection.observe(select_analysis_eventhandler, names='value')
    
    export_result_button = widgets.Button(description="Export Result to d4.out", icon="download", layout={'width': '250px'}, disabled=False)
    def export_result_button_eventhandler(change):
        nonlocal anl_result
        nonlocal anl_out
        if anl_result is not None:
            d4.out = anl_result
            with anl_out:
                display(Markdown("_d4.out variable updated with analisys result_"))
    export_result_button.on_click(export_result_button_eventhandler)
    
    first_options = widgets.Box([object_selection, type_selection, dataframe_selection])
    second_options = widgets.Box([analysis_selection, export_result_button])
    if is_df:
        analysis_selection.options = get_analysis_options(anl_src)
        if len(analysis_selection.options) == 1:
            anl_out.clear_output()
            with anl_out:
                display(Markdown("**No analysis available**"))
        else:
            second_options.layout.visibility='visible'
    else:
        second_options.layout.visibility='hidden'
    # anl_out.clear_output()
    display(Markdown("**Analisys explorer:**"))
    display(first_options)
    display(second_options)
    display(anl_out)

# gui short alias functions =============================================================
def xmn(*args, **kwargs):
    """ Alias for xmenu()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return xmenu(*args, **kwargs) 

def xdp(*args, **kwargs):
    """ Alias for xdisplay()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return xdisplay(*args, **kwargs) 

def rdg(*args, **kwargs):
    """ Alias for read_data_gui()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return xread(*args, **kwargs) 



# DS4N6
#
# Description: library of functions to appy Data Science in several forensics
#              artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4fstl

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import inspect

import xmltodict
import json
import pickle
from tqdm import tqdm
import xml.etree.ElementTree as et

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import display, Markdown, HTML

from sklearn.model_selection import train_test_split
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.layers import Input, Dense

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui
import ds4n6_lib.utils  as d4utl

###############################################################################
# IDEAS
###############################################################################
# is_deleted()
# is_file()
# is_dir() / is_folder() - level
# ext() # filter by Extension
# nofn  # exclude $FILE_NAME entries

###############################################################################
# FUNCTIONS
###############################################################################
# FILE READING FUNCTIONS ######################################################

def read_data(evdl, **kwargs):
    return d4com.read_data_common(evdl, **kwargs)
    
# CORE FUNCTIONS (simple, analysis, etc.) #####################################
# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """
    if d4.debug >= 4:
        print("DEBUG: [fstl] [simple_func()]")

    # Variables ----------------------------------------------------------------
    hiddencols =  []

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Call to simple_common ----------------------------------------------------
    return d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)

# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4fstl")
class Ds4n6FSTLAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)

    def nofn(self):
        return self._obj[~self._obj['FileName'].str.contains(r"\ \(\$FILE_NAME\)")]
        
    def is_deleted(self):
        if 'FileName' in self._obj.columns:
            return self._obj[self._obj['FileName'].str.contains(r"\ \(deleted\)$")]
        elif 'Deleted_' in self._obj.columns:
            return self._obj.query('Deleted_ == True')
        
    def is_file(self):
        return self._obj.query('Type_ == "r" | PrevType_ == "r"')
        
    def is_dir(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    # Same as is_dir()
    def is_directory(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    # Same as is_dir()
    def is_folder(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    def ext(self,ext):
        return self._obj[self._obj['FileName'].str.contains(r"\."+ext+"$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(\$FILE_NAME\)$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(deleted\)$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(\$FILE_NAME\)\ \(deleted\)$")]

    def just_basename(self):
        return self._obj['FileName'].str.replace('.*/','')
        
    def ts_m(self,exclusive=False):
        if exclusive == False:
            return self._obj.query(r'MACB.str.contains("^m...$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^m\.\.\.$")',engine="python")
        
    def ts_a(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^.a..$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.a\.\.$")',engine="python")
        
    def ts_c(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^..c.$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.\.c\.$")',engine="python")
        
    def ts_b(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^...b$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.\.\.b$")',engine="python")
   
@pd.api.extensions.register_dataframe_accessor("d4_fstl")
class Ds4n6_FSTLAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)

    def nofn(self):
        return self._obj[~self._obj['FileName'].str.contains(r"\ \(\$FILE_NAME\)")]
        
    def is_deleted(self):
        if 'FileName' in self._obj.columns:
            return self._obj[self._obj['FileName'].str.contains(r"\ \(deleted\)$")]
        elif 'Deleted_' in self._obj.columns:
            return self._obj.query('Deleted_ == True')
        
    def is_file(self):
        return self._obj.query('Type_ == "r" | PrevType_ == "r"')
        
    def is_dir(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    # Same as is_dir()
    def is_directory(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    # Same as is_dir()
    def is_folder(self,level=0):
        return self._obj.query('Type_ == "d"')
        
    def ext(self,ext):
        return self._obj[self._obj['FileName'].str.contains(r"\."+ext+"$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(\$FILE_NAME\)$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(deleted\)$") | self._obj['FileName'].str.contains(r"\."+ext+r"\ \(\$FILE_NAME\)\ \(deleted\)$")]

    def just_basename(self):
        return self._obj['FileName'].str.replace('.*/','')
        
    def ts_m(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^m...$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^m\.\.\.$")',engine="python")
        
    def ts_a(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^.a..$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.a\.\.$")',engine="python")
        
    def ts_c(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^..c.$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.\.c\.$")',engine="python")
        
    def ts_b(self,exclusive=False):
        if exclusive == False:
            return self._obj.query('MACB.str.contains("^...b$")',engine="python")
        else:
            return self._obj.query(r'MACB.str.contains("^\.\.\.b$")',engine="python")
   
# ANALYSIS ####################################################################

# analysis() function =========================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    return analysis_func(*args, **kwargs)

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """

    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        # anlav = False
        print("Available fstl analysis types:")
        print("- No analysis functions defined yet.")
        return

        # TEMPLATE
        #if objtype == "str-help" or objtype == "str-list" or  re.search("^pandas_dataframe-fstl-mactime-standard", objtype):
        #    anlav = True
        #    print("- XXXXXXXXXX:  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX (Input: fstldf)")

        # if anlav == False:
        #     print('- No analysis modules available for this object ('+objtype+').')

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # TEMPLATE
    # If object is a dict of dfs
    #elif re.search("^pandas_dataframe-evtx_file_df", objtype):
    #    if anltype == "XXXXXXXXXXX":
    #        return XXXXXXXXXXXXXXXXXXXXX(*args, **kwargs)
    #else:
    #    print("ERROR: [fstl] Unsupported input data.")
    #    return





# DS4N6
#
# Description: Library of functions to apply Data Science to forensics artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4flst

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import pickle
import inspect

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui
import ds4n6_lib.utils  as d4utl
import ds4n6_lib.unx    as d4unx

###############################################################################
# VARIABLES
###############################################################################
hiddencols = [ 'MTStampEpoch_', 'MTStamp_', 'ATStampEpoch_', 'ATStamp_', 'CTStampEpoch_', 'CTStamp_', 'Meta_', 'FileStem_', 'ParentName_', 'ParentPath_', 'ParentMeta_', 'PathSeparator_', 'FilePath-Hash_', 'FileName-Hash_', 'FileStem-Hash_', 'ParentPath-Hash_', 'ParentName-Hash_', 'NTFS-SeqNumber_', 'ParentSeqNumber_', 'ParentPath', 'NTFS-ReferenceCount_', 'NTFS-ReparseTarget_', 'IsDirectory_', 'NTFS-HasAds_', 'NTFS-IsAds_', 'NTFS-SI<FN_', 'NTFS-uSecZeros_', 'NTFS-Copied_', 'NTFS-SiFlags_', 'NTFS-NameType_', 'NTFS-FN-BTime_', 'NTFS-FN-MTime_', 'NTFS-FN-CTime_', 'NTFS-FN-ATime_', 'NTFS-UpdateSequenceNumber_', 'NTFS-LogfileSequenceNumber_', 'NTFS-SecurityId_', 'NTFS-ObjectIdFileDroid_', 'NTFS-LoggedUtilStream_', 'NTFS-ZoneIdContents_', ]

###############################################################################
# FUNCTIONS
###############################################################################

# ANALYSIS FUNCTIONS ##########################################################

# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """

    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Artifact-specific argument parsing =======================================

    # Variables ================================================================
    dfout = df

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Pre-Processing ==========================================================

    # Call to simple_common ===================================================
    dfout = d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)

    # Post-Processing =========================================================

    # Return ==================================================================
    return dfout

# analysis ====================================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return analysis_func(*args, **kwargs)

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """

    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        anlav = False
        print("Available flist analysis types:")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.search("^pandas_dataframe-flist", objtype):
            anlav = True
            print("- size_top_n:  Top files by size (Input: flistdf)")
            print("- exefile:  Analysis of the multiple instances of a specific file on many hosts (Input: flistdf)")
            print("- unique_files_folder:  Find unique files (Input: flistdf)")
            print("- exefs:  Macro analysis of all EXEs in the exefs df as a whole (Input: flistdf)")

        if anlav == False:
            print('- No analysis modules available for this object ('+objtype+').')

    if d4.debug >= 2:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    nargs = len(args)
    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]
    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # ANALYSIS FUNCTIONS ======================================================

    # flistdf ------------------------------------------------------------------
    if   re.search("^pandas_dataframe-flist", objtype):
        if anltype == "size_top_n":
            return analysis_size_top_n(obj)
        if anltype == "exefile":
            return analysis_exefile(obj)
        if anltype == "unique_files_folder":
            return analysis_unique_files_folder(obj)
        if anltype == "exefs":
            return analysis_exefs(obj)

    print("INFO: [d4flist] No analysis functions available for this data type ("+objtype+")")

# Specific Analysis Functions =================================================

def analysis_size_top_n(fstl, n=100):
    return fstl[~fstl['FileName_'].str.contains(r"\(\$FILE_NAME\)")][['Size_','FileName_']].sort_values(by='Size_', ascending=False).drop_duplicates().head(n)


def analysis_exefile(exefs, thisexef_path='notepad.exe'):
    # Description:
    #     Analysis of the multiple instances of a specific file on many hosts
    #     - ...

    # Variables
    exef_intg_max_occs = 3
    print("PARAMETERS --------------------------------------------------------\n")
    print("exe path: {}".format(thisexef_path))
    print("EXEFILE ANALYSIS --------------------------------------------------------\n")

    # Select instances of exef
    exefs[exefs['FilePath_'].str.contains(thisexef_path,case=False)].head()
    thisexefs = exefs[exefs['FilePath_'].str.contains(thisexef_path,case=False)]

    print("fsize   ANALYSIS - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n")
    exefgrps = thisexefs.groupby('Size_')
    exefgrps_groups = exefgrps.groups
    nexefgrps = len(exefgrps_groups)
    print("No.groups: " + str(nexefgrps) + "\n")
    # exef_sizes = exefgrps.groups.keys()
    exef_sizes_occs = exefgrps.size()
    print("Groups (sorted by no. occurrances of 'fsize'):    ")
    print(exef_sizes_occs.sort_values(ascending=False))

    print("Interesting (no. occurrences <=" + str(exef_intg_max_occs) + "):    ")
    exef_intg=exefgrps.filter(lambda x: len(x) <= exef_intg_max_occs)
    print(exef_intg)
    return exef_intg

    
def analysis_unique_files_folder(exefs, thisexed_path="/Windows/Temp", exef_intg_max_occs="100", compop='==', recurse=False, prevdays=0, tsfield='m', verbose=False):
    print("PARAMETERS --------------------------------------------------------\n")
    print("exe path: {}".format(thisexed_path))
    print("max occs: {}".format(exef_intg_max_occs))
    print("compop: {}".format(compop))
    print("recurse: {}".format(recurse))
    print("prevdays: {}".format(prevdays))
    print("verbose: {}".format(verbose))
    print("-------------------------------------------------------------------\n")
    if compop not in ['>', '<', '>=', '==', '<=']:
        print("Invalid Comparison Operator: "+compop)
        return False

    regexrec=thisexed_path+"/"
    regexnorec=thisexed_path+"/[^/]*$"

    if recurse :
        thisexefsrec=exefs[exefs['FilePath_'].str.contains(regexrec,case=False,regex=True)]
        nexefsrec=len(thisexefsrec)
        thisexefs=thisexefsrec
        if verbose :
            print("No. files (recursive):     "+str(nexefsrec)+"\n")
    else:
        thisexefsnorec=exefs[exefs['FilePath_'].str.contains(regexnorec,case=False,regex=True)]
        nexefsnorec=len(thisexefsnorec)
        thisexefs=thisexefsnorec
        if verbose :
            print("No. files (non-recursive): "+str(nexefsnorec)+"\n")

    exefgrps = thisexefs.groupby('FilePath_')
    exefgrps_groups = exefgrps.groups
    nexefgrps = len(exefgrps_groups)
    # exef_sizes=exefgrps.groups.keys()
    # exef_sizes_occs=exefgrps.size()
    if verbose :
        print("phash ANALYSIS - - - - - - - - - - - - - - - - - - - - - - - - - - - - - \n")
        print("RECURSION: "+str(recurse))
        print("No.groups: "+str(nexefgrps)+"\n")

    if prevdays == 0 :
        exef_intg = exefgrps.filter(lambda x: eval( str(len(x)) + compop + str(exef_intg_max_occs)) )
    else:
        print("No. Interesting (no. occurrences <=" + str(exef_intg_max_occs) + "): " + str(exef_intg) + "\n")
        lastmtime = exef_intg.sort_values(by="MTime_").tail(1)['MTime_']
        print("Last mtime: " + str(lastmtime))
        prevdate = lastmtime + pd.DateOffset(days=-prevdays)
        print("Previous Date: "+  str(prevdate))

    return exef_intg


def analysis_exefs(exefs,  thisexed_path="/Windows/Temp", date_from='2019-12-01'):

    # Description:
    #     Macro analysis of all EXEs in the exefs df as a whole
    #     - Files that appear only a few times, etc.
    print("PARAMETERS --------------------------------------------------------\n")
    print("date from: {}".format(date_from))
    # path-hash analysis  ===============================================================
    # rare_phash_occs = 3
    # thisexefilegrps = thisexefs.groupby('path-hash')
    # thisexefilegrps_groups = thisexefilegrps.groups
    # nthisexefilegrps = len(thisexefilegrps_groups)
    # thisexefile_phash = thisexefilegrps.groups.keys()
    # thisexefile_phash_occs = thisexefilegrps.size()
    # print("Groups (sorted by no. occurrances of 'path-hash'):    ")
    # print(thisexefile_phash_occs.sort_values(ascending=False))
    # thisexefile_phash_rare = thisexefile_phash_occs[thisexefile_phash_occs == 1].sort_values(ascending=False)

    # Rare Files Analysis ---------------------------------------------------------------
    # Files which appear only n times in the whole host set. 
    # Since we are grouping by phash, this means they appear once in each of the n hosts
    # If n=1 -> file appears only in 1 host
    # exef_intg = exefgrps.filter(lambda x: len(x) <= n)    

    # Creation time analysis ------------------------------------------------------------
    # Files created recently
    return exefs[exefs['ATime_'] > date_from]
# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4fsl")
class Ds4n6FslAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)


@pd.api.extensions.register_dataframe_accessor("d4_flist")
class Ds4n6FSListAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)



# DS4N6
#
# Description: library of functions to appy Data Science in several forensics
#              artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4ksa

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import inspect
import xmltodict
import json
import pickle
from tqdm import tqdm
import xml.etree.ElementTree as et

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import display, Markdown, HTML

from sklearn.model_selection import train_test_split
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.layers import Input, Dense

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4      as d4
import ds4n6_lib.common  as d4com
import ds4n6_lib.gui     as d4gui
import ds4n6_lib.utils   as d4utl

###############################################################################
# FUNCTIONS
###############################################################################

# FILE READING FUNCTIONS ######################################################

def read_data(evdl, **kwargs):
    return d4com.read_data_common(evdl, **kwargs)

# HARMONIZATION FUNCTIONS #####################################################

def harmonize(df, **kwargs):
    """ Convert DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """

    # (1) kansa will probably be invoked with 'tool=kansa', but kansa is in 
    #     reality an orchestrator, so we will only populate the D4_Tool_ column 
    #     to kansa only if we have not been able to determine what is the 
    #     underlying tool that kansa is using for execution in the endpoint

    # Specific Harmonization Pre-Processing -----------------------------------
    if not 'D4_Orchestrator_' in df.columns:
        df.insert(0, 'D4_Orchestrator_', "kansa")
    else:
        # If the D4_Orchestrator_ col exists, we are in a recursive call
        return df

    objtype = d4com.data_identify(df)

    # Generic Harmonization ---------------------------------------------------

    # Since kansa is an orchestrator, let's try to identify the specific 
    # data type and apply the corresponding harmonization function.
    # If we can, we will execute the generic one.
    if "unknown" in objtype:
        df = d4com.harmonize_common(df, **kwargs)
    else:
        # Let's try to harmonize this specific df
        # WARNING: Since we no longer identify datatype by DF cols, this will
        #          not work
        df = d4com.harmonize(df)

    # Specific Harmonization Post-Processing ----------------------------------
    df['D4_Hostname_'] = df['PSComputerName']

    if df['D4_Plugin_'].iloc[0] == "Tasklistv":
        df['D4_DataType_'] = "pslist"
        df['D4_DataType_'] = df['D4_DataType_'].astype('category')

        # Rename columns
        df = df.rename(columns={'ImageName': 'Name_', 'PID': 'PID_', 
                                'SessionName': 'SessionName_', 
                                'SessionNum': 'Session_', 'MemUsage': 'MemUsage_', 
                                'Status': 'Status_', 'UserName': 'UserName_', 
                                'CPUTime': 'CPUTime_', 'WindowTitle': 'WindowTitle_'
                               })

    elif df['D4_Plugin_'].iloc[0] == "SvcAll":
        df['D4_DataType_'] = "svclist"
        df['D4_DataType_'] = df['D4_DataType_'].astype('category')

        # Rename columns
        df = df.rename(columns={'Name': 'Name_', 'DisplayName': 'DisplayName_', 
                                'PathName': 'FilePath_', 'StartName': 'UserName_',
                                'StartMode': 'StartMode_', 'State': 'State_',
                                'TotalSessions': 'TotalSessions_', 
                                'Description': 'Description_'
                               })

    return df

# ANALYSIS FUNCTIONS ======================================================

# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """
    if d4.debug >= 4:
        print("DEBUG: [ksa] [simple_func()]")

    # Variables ----------------------------------------------------------------
    hiddencols =  []

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Call to simple_common ----------------------------------------------------
    return d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)

# analysis ====================================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    return analysis_func(*args, **kwargs)

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """
    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        anlav = False
        print("Available kansa analysis types:")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.search("^dict-pandas_dataframe-kansa", objtype):
            anlav = True
            print("- kansa_files:  No.events kansa file (Input: ksadfs)")

        if anlav == False:
            print('- No analysis modules available for this object ('+objtype+').')

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # ksadfs ------------------------------------------------------------------
    if   re.search("^dict-pandas_dataframe-kansa", objtype):
        if anltype == "kansa_files":
            return analysis_kansa_files(*args, **kwargs)

    print("INFO: [d4ksa] No analysis functions available for this data type ("+objtype+")")

def analysis_kansa_files(*args, **kwargs):
    """ Analysis that gives kansa files

        Args: 
        obj:          Input data (typically DF or dict of DFs)
        Returns: 
        pandas.Dataframe with the results of the analysis

    """
    dfs = args[0]

    objtype = d4com.data_identify(dfs)

    if objtype != "dict-pandas_dataframe-kansa":
        print("ERROR: Invalid object for function: "+objtype)
        print("       Input object should be:      dict-pandas_dataframe-kansa")
        return

    outdf = pd.DataFrame([],columns=['File','NEntries'])
    row = pd.Series()

    for key in dfs.keys():
        row['File']  = key
        row['NEntries'] = len(dfs[key])

        outdf = outdf.append(row,ignore_index=True)

    return outdf

# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4ksa")
class Ds4n6KsaAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)

@pd.api.extensions.register_dataframe_accessor("d4_kansa")
class Ds4n6KansaAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)





import os
import glob
from tensorflow.keras import layers
from tensorflow.keras.models import Model
import tensorflow as tf
from tensorflow.keras import losses
import matplotlib.pyplot as plt
import pandas as pd


def convert_prefetch_ham_to_hml(df):
    df_split = df['file_referenced'].str.split("\\",expand=True)
    df_split = df_split.drop(columns=[0]).fillna(value='')
    
    first_column = df_split.iloc[:, 0]
    medium_column = []
    last_column = []
    for i in range(df_split.shape[0]):
        arr = [x for x in df_split.iloc[i, 1:] if x != '']
        medium_column.append('\\'.join(arr[:-1]))
        last_column.append('\\'.join(arr[-1:])) # [-1:] because some len(arr) == 0 
    
    list_to_df = list(zip(first_column, medium_column, last_column, df['machine_id']))
    new_df = pd.DataFrame(list_to_df, columns =['A', 'B', 'C', 'machine_id'])
    return new_df


def ml_prefetch_anomalies(df, odalg="simple_autoencoder", latent_dim = 128, epochs = 10, learning_rate = 1e-3):    
    # Deep Learning
    x_train = pd.get_dummies(df).to_numpy()
    
    class Autoencoder(Model):
      def __init__(self, input_dim, latent_dim):
        super(Autoencoder, self).__init__()
        self.input_dim = input_dim
        self.latent_dim = latent_dim   
        self.encoder = layers.Dense(latent_dim, activation='relu')
        self.decoder = layers.Dense(input_dim, activation='sigmoid')

      def call(self, x):
        encoded = self.encoder(x)
        decoded = self.decoder(encoded)
        return decoded

    def train_autoencoder(latent_dim, epochs, learning_rate):
        autoencoder = Autoencoder(input_dim=x_train.shape[1], latent_dim=latent_dim)
        opt = tf.keras.optimizers.Adam(learning_rate=learning_rate)
        autoencoder.compile(optimizer=opt, loss=losses.MeanSquaredError())
        history = autoencoder.fit(x_train, x_train, epochs=epochs, shuffle=True, verbose=0)
        return autoencoder, history

    model, history = train_autoencoder(latent_dim=latent_dim,
                                       epochs=epochs,
                                       learning_rate=learning_rate)


    preds = model.predict(x_train)
    inference_losses = tf.keras.metrics.mean_squared_error(preds, x_train.astype('float')).numpy()
    
    ranking = []
    for i, loss in zip(range(len(inference_losses)), inference_losses):
        fr = '\\'.join(df.iloc[i, :3])
        
        machine_id = df.iloc[i]['machine_id']
        if fr.endswith('.DLL'):
            ranking.append((loss, i, fr, machine_id))

    ranking = sorted(ranking, key=lambda x: -x[0])
    anomdf = pd.DataFrame(ranking, columns=['loss', 'source_index', 'file referenced', 'machine_id'])
    return anomdf[['file referenced', 'machine_id']]



# DS4N6
#
# Description: library of functions to appy Data Science in several forensics
#              artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4pl

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import inspect
import xmltodict
import json
import pickle
from tqdm import tqdm
import xml.etree.ElementTree as et
import subprocess
import datetime

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt
import ipywidgets as widgets

from IPython         import get_ipython
from IPython.display import display, Markdown, HTML, Javascript

# ML  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
from sklearn.model_selection import train_test_split

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.utils  as d4utl
import ds4n6_lib.evtx   as d4evtx
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui

###############################################################################
# FUNCTIONS
###############################################################################

# FILE READING FUNCTIONS ######################################################

def read_data(evdl, **kwargs):
    """ Read data from files or a folder

        Args: 
            evdl (str): path to file/folder source
            kwargs: read options
        Returns: 
            pandas.Dataframe or dictionary of pandas.DataFrame
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    kwargs.setdefault('tool', 'plaso')

    if   bool(re.search(r"\.csv$", evdl, re.IGNORECASE)):
        return read_plaso_l2tcsv(evdl, **kwargs)
    
    elif bool(re.search(r"\.json$", evdl, re.IGNORECASE)):
        return read_plaso_json(evdl, **kwargs)
    else:
        print("ERROR: Unable to read input file. Unsupported file extension.")
        return

def read_plaso_l2tcsv(plasof, **kwargs):
    """ Read plaso data from from a l2tcsv file

        Args: 
            plasof (str): path to file source
            kwargs: read options
        Returns: 
            dictionary of pandas.DataFrame
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    
    dfs = {}
    # When creating the .csv file with psort, make sure to filter out events 
    # with absurd timestamps (e.g. < 1970 or > the current year). 
    # plaso sometimes includes "broken" timestamps and the will break pandas
    # psort.py -z UTC -o l2tcsv -w plaso_file.csv plaso_file.dump \ 
    #     "date > '1960-01-01 00:00:00' AND date < '2020-12-31 23:59:59'"
    plines=pd.read_csv(plasof)
    
    if 'date' in plines.columns:
        # Now, we need to clean the data. There may be wrong values.
        plines=plines[~plines['date'].str.contains('/00/')]
        # Let's include:
        # - a timestamp column which combines the date and time cols, and is 
        #   of type datetime64
        # - a column "sourcetype_clean" similar to "sourcetype" but "cleaner",
        #   i.e. easier to use
        plines.insert(0, 'timestamp', 0)
        plines.insert(7, 'sourcetype_clean', '')
        plines['sourcetype_clean']=plines['sourcetype'].str.replace('UNKNOWN : ','').str.replace(' ','_')
        plines['timestamp']=pd.to_datetime(plines['date']+" "+plines['time'])

        # Let's create a DF for each source (pe_compilation_time, winlogon, ...),
        # It will be easier for analysis
        srctypes = plines['sourcetype_clean'].unique()
        for srctype in srctypes:
            srctypevar = str(srctype).lower()
            print('Reading source_type %-40s into dataframe ->  %-40s' % (srctype,srctypevar))
            dfs[srctypevar] = plines[plines['sourcetype_clean'] == srctype ].copy()
    
    else:
        dfs = plines
    
    
    return dfs

def read_plaso_json(evdl, **kwargs):
    """ Read plaso data from from a json file

        Args: 
            evdl (str): path to file source
            kwargs: read options
        Returns: 
            dictionary of pandas.DataFrame
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    #
    # Generate your plaso json file with the following command:
    #     psort.py -z UTC -o json -w plaso_file.json plaso_file.dump "date > '1960-01-01 00:00:00' AND date < '2021-01-01 00:00:00'"

    # Parse Arguments
    tool                     = kwargs.get('tool',                    '')
    hostname                 = kwargs.get('hostname',                '')
    do_harmonize             = kwargs.get('harmonize',               True)
    build_all_df             = kwargs.get('build_all_df'            , False)
    use_pickle               = kwargs.get('use_pickle'              , True)

    pklrawf = evdl+'.raw.pkl'
    pklhtmf = evdl+'.ham.pkl'

    if os.path.exists(pklhtmf) and use_pickle  and do_harmonize :

        # Read from pickle
        print("- Saved Harmonized pickle file found:")
        print("      "+pklhtmf)
        print("- Reading data from HAM pickle file...")
        dfs = pickle.load(open(pklhtmf, "rb"))
        print("- Done.")
        print("")

    else:
        print("- No saved Harmonized pickle file found.")
        print("")
    
        # Read Data - - - - - - - - - - - - - - - - - - - - - - - - - -
        if os.path.exists(pklrawf) and use_pickle :
            print("- Saved RAW pickle file found:")
            print("      "+pklrawf)
            print("- Reading data from RAW pickle file...")
            dfs = pickle.load(open(pklrawf, "rb"))
            print("- Done.")
            print("")
        else:
            if not os.path.exists(evdl):
                print("ERROR: File does not exist:")
                print("         "+evdl)
                print("")
                return

            dfs = {}
            print("    " + evdl)
            print('- Reading json file lines: ',end='')
            cnt = 0
            with open(evdl) as infile:
                evtsdict = {}
                for line in infile:
                    if line.strip() != "}":
                        lineclean = re.sub(r'^[,{]\ *','', line)
                        datatype = lineclean.split('"data_type": ', maxsplit=1)[-1].split(maxsplit=1)[0].replace('"', '').replace(",", "").replace(":","_")
            
                        if datatype not in evtsdict.keys():
                            evtsdict[datatype]={}
                        linecleanjson = json.loads("{"+lineclean+"}")
                        evtsdict[datatype].update(linecleanjson)

                    # Print progress counter
                    if cnt % 50000 == 0:
                        print("[" + str(cnt) + "] ", end='')
                    cnt=cnt+1
            
            print("\n")
            print("- Generating pandas dataframes: ")
            keys = evtsdict.keys()
            for key in keys:
                # Convert json string to pandas DF
                nevtskey = len(evtsdict[key])
                print('  + %-45s ... ' % (key), end='')
                df = pd.DataFrame(evtsdict[key]).T

                dfs[key] = df

                print(' [%s]' % (str(nevtskey)))

        # Harmonize - - - - - - - - - - - - - - - - - - - - - - - - - -
        if do_harmonize :
            print(" ")
            print("- Harmonizing pandas dataframes: ")
            keys = dfs.keys()
            for key in keys:
                print('  + %-45s ... ' % (key), end='')
                if  len(dfs[key]) > 0:
                    dfs[key] = harmonize(dfs[key], tool=tool, plugin=key, hostname=hostname)
                    print(' [OK]')
                else:
                    print(' [EMPTY]')

            if build_all_df :
                dfs['all'] = pd.concat([dfs['all'], dfs[key]], ignore_index=True)

        print("- Done Harmonizing.\n")

        # Save to pickle - - - - - - - - - - - - - - - - - - - - - - - - - -
        d4com.save_pickle(evdl, dfs, use_pickle, do_harmonize)
        

    print("\nNOTE: Now you can use the syntax <yourvar>['<datatype>'] to access your specific dataframe")

    return dfs

# HARMONIZATION FUNCTIONS #####################################################

# Tool Output Harmonization -> HTM
def harmonize(df, **kwargs):
    """ Convert DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Specific Harmonization Pre-Processing =================================== 

    # Name index as pevtnum (plaso evt number) and reset it (as a std col)
    df.index = df.index.set_names("pevtnum")
    df = df.reset_index()

    # Resort columns
    cols = df.columns.tolist()
    if 'message' in df.columns:
        cols.insert(len(cols), cols.pop(cols.index('message')))
    cols.insert(len(cols), cols.pop(cols.index('pevtnum')))
    cols.insert(0, cols.pop(cols.index('timestamp_desc')))
    df = df.reindex(columns=cols)

    df['timestamp_desc'] = df['timestamp_desc'].str.replace(' Time$','').str.replace('Modification','Modif.')

    # Generic Harmonization ===================================================
    df = d4com.harmonize_common(df, **kwargs)

    # Specific Harmonization Post-Processing ==================================

    # Convert timestamp to datetime
    nowus = int(time.time()*1000000)
    df.insert(5, 'Timestamp_', None)
    df['Timestamp_'] = df['timestamp']
    df['Timestamp_'] = df['Timestamp_'].mask(df['Timestamp_'].lt(0),0)
    df['Timestamp_'] = df['Timestamp_'].mask(df['Timestamp_'].gt(nowus),0)
    df['Timestamp_'] = pd.to_datetime(df['Timestamp_'], unit="us")

    # Resort columns
    cols = df.columns.tolist()

    # return ==================================================================
    return df

def harmonize_plaso_artifact_common(df):
    """ Common part of the proccess Convert plaso DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")
    cols = ['timestamp_desc', '__container_type__', '__type__', 'data_type', 'inode', 'parser', 'pathspec', 'sha256_hash', 'message', 'pevtnum'] 
    df = df.drop(columns=[col for col in cols if col in df.columns])
    return df

# ARTIFACT EXTRACTION FUNCTIONS ###############################################

# Index:
# - evtx
# - windows_registry_key_value
# - windows_registry_service
# - amcache 
#   + windows_registry_amcache 
#   + windows_registry_amcache_programs

# evtx ========================================================================
def plaso_extract_single_evtx(plevtxdf, evtxf, nwf=d4.main_nwf, recovered=False, save_xml=False, xml_filename=""):

    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    if recovered :
        thisplevtxdf = plevtxdf.query('LogFile == @evtxf')
    else:
        thisplevtxdf = plevtxdf.query('LogFile == @evtxf and recovered == False')

    # Extract xml_string field
    print("    - Extracting & concatenating plaso xml_string entries")
    xmlstr = thisplevtxdf['xml_string'].str.cat()
    xmlstr = xmlstr.replace('\\n',' ')

    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    # nwf - XML not well-formed
    if nwf != "":
        print("    - Escaping/fixing not well-formed records")
        for rule in nwf:
            next = False
            
            if "file" in rule:
                if rule.get('file') != evtxf:
                    next = True
                else:
                    print("  Rule for this evtx: %s" % evtxf)
            
            if next == False:
                thisstart = len(xmlstr)
                if rule.get('type') == "re":
                    xmlstr = re.sub(rule.get('find'), rule.get('replace'), xmlstr)
                else:
                    xmlstr = xmlstr.replace(rule.get('find'),rule.get('replace'))

                if thisstart != len(xmlstr):
                    print("      + Rule executed: %s" % rule.get('replace'))
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    xmlstr='<Events>'+xmlstr+'</Events>'

    if save_xml :
        if xml_filename == "":
            xml_filename = "plaso_extracted_evtx_events.xml"

        print("    - Saving xml to file:")
        print("        "+xml_filename)

        xmlfile = open(xml_filename, "w")
        xmlfile.write(xmlstr)
        xmlfile.close()

    # We no longer need plevtxdf. Let's clear it to save memory
    plevtxdf = pd.DataFrame([])

    try:
        print("    - Converting XML -> DF")
        outdf = d4utl.xml_to_df(xmlstr, sep=" > ").d4evtx.column_types_set()
        print("")
    except:
        print("    - ERROR: Could not convert XML -> DF")
        outdf = pd.DataFrame()
        print("")

    return outdf

def plaso_extract_evtx(plobj, nwf=d4.main_nwf, recovered=False, save_xml=False, xml_filename="",files2process=[]):
    # Input can be either a full plaso DFs dict or a DF of plaso evtx

    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    evtxdfs={}

    if isinstance(plobj, dict):
        plevtxdf = plobj['windows_evtx_record']
    elif isinstance(plobj, pd.DataFrame):
        plevtxdf = plobj
    else:
        print("ERROR: Unsupported Input Data")
        return
 
    # Remove the path / preserve the file name only
    # OLD:
    #  # Windows path (if plaso has been run on a disk image or live Windows System)
    #  plevtxdf['LogFile'] = plevtxdf['pathspec'].apply(lambda x: x['location']).str.replace(".*\\\\","").copy()
    #  # Linux path (if plaso has been run on a Linux System, e.g. on extracted Windows artifacts)
    #  plevtxdf['LogFile'] = plevtxdf['pathspec'].apply(lambda x: x['location']).str.replace(".*/","").copy()
    # 

    # For evtx, since they are self-contained artifacts, we can use the plaso "filename" field 
    # Windows path (if plaso has been run on a disk image or live Windows System)
    plevtxdf['LogFile'] = plevtxdf['filename'].str.replace(".*\\\\","").copy()
    # Linux path (if plaso has been run on a Linux System, e.g. on extracted Windows artifacts)
    plevtxdf['LogFile'] = plevtxdf['filename'].str.replace(".*/","").copy()

    #plevtxdf['LogFile'] = plevtxdf['pathspec'].apply(lambda x: "["+str(x['parent']['location'])+"]"+str(x['location']))
    #evtxfs = plevtxdf['LogFile'].drop_duplicates() 
    #evtxfs=plevtxdf['pathspec'].apply(lambda x: "["+str(x['parent']['location'])+"]"+str(x['location'])).str.replace("/","",1).drop_duplicates().str.replace(".*\\\\","")

    # Event Log files present in the plaso evtx DF
    #evtxfs=plevtxdf['pathspec'].apply(lambda x: "["+str(x['parent']['location'])+"]"+str(x['location'])).str.replace("/","",1).drop_duplicates().str.replace(".*\\\\","")

    # Include the "recovered" column from plaso
    plevtxdf['recovered'] = plevtxdf['recovered'].astype(bool)

    # DICT OF INDIVIDUAL DFs (one per evtx log file) - - - - - - - - - - - - - - - - - - - 
    if files2process != []:
        # TODO: This has not been tested yet. Chances are that this doesn't work and you 
        #       need to polish this 
        evtxfs = files2process
        print("- Files to process:   " + str(files2process))
        print("\n")
    else:
        evtxfs = plevtxdf['LogFile'].drop_duplicates() 
        files2process = evtxfs.to_list()

    print("- Files to process: " + str(files2process))

    print("- Processing:")

    for evtxf in evtxfs:
        evtxfbase = re.sub('^.*\\\\', '', evtxf)
        print("  + " + evtxfbase, end='')

        # If the files2process arg is provided, process just the logs specified
        if evtxf not in files2process:
            print(" [Skipping]")
            continue
        else:
            print("\n")

        xml_filename_file=None
        # XML Filename
        if save_xml :
            xml_filename_base = re.sub(r"\.xml$", "", xml_filename)
            xml_filename_file = xml_filename_base + "-" + evtxfbase + ".xml"

        evtxdf = plaso_extract_single_evtx(plevtxdf, evtxf, nwf, recovered, save_xml=save_xml, xml_filename=xml_filename_file)

        orchestrator = plevtxdf['D4_Orchestrator_'].iloc[0]
        tool         = plevtxdf['D4_Tool_'].iloc[0]
        plugin       = plevtxdf['D4_Plugin_'].iloc[0]
        hostname     = plevtxdf['D4_Hostname_'].iloc[0]
        datatype     = "evtx-raw"
        
        if not 'evtxFileName_' in evtxdf.columns:
            evtxdf.insert(0, 'evtxFileName_', evtxfbase)

        # Insert Tool / Data related HAM Columns
        if not 'D4_Hostname_' in evtxdf.columns:
            evtxdf.insert(0, 'D4_Hostname_', hostname)
        if not 'D4_Plugin_' in evtxdf.columns:
            evtxdf.insert(0, 'D4_Plugin_', plugin)
        if not 'D4_Tool_' in evtxdf.columns:
            evtxdf.insert(0, 'D4_Tool_', tool)
        if not 'D4_Orchestrator_' in evtxdf.columns:
            evtxdf.insert(0, 'D4_Orchestrator_', orchestrator)
        if not 'D4_DataType_' in evtxdf.columns:
            evtxdf.insert(0, 'D4_DataType_', datatype)

        evtxdfs[evtxfbase] = evtxdf

        print(" [OK] ")

    print("- Done")

    return evtxdfs

                
def plaso_get_evtxdfs(pldfs, hostname, datapath="", filename="", notebook_file="", evtxdfs_usepickle=True, save_xml=False, xml_filename="",files2process=[], savepathtonb=True):

    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    filenamebase = os.path.basename(filename)

    print("Extracting evtx DFs dict from plaso DFs dict.")

    if notebook_file != "":
        
        print('- Looking for notebook cell containing pickle file to read ('+hostname+'_evtxdfspklf)')
        hits = d4utl.nbgrep(notebook_file, "# "+hostname+'_evtxdfspklf_f2read = "/.*$')

        if hits:
            f2read = hits[0].split(" = ")[1].strip('"')
            print("  + Found: "+f2read)
        else:
            print("  + Not found")
            f2read = ""
    else:
        f2read = ""
        
    if f2read != "":
        pklinnb = True
        evtxdfspklf = f2read
    else:
        pklinnb = False
        # If we don't find the cell with the saved path for evtxdfspklf,
        # we will try to find the file in the datapath directory
        if datapath != "":
            if filename == "":
                file2read = hostname + '.evtxdfs.raw.pickle'
            else:
                file2read = filenamebase
                
            evtxdfspklf = datapath + '/' + file2read
            print("- Looking for evtxdfs pickle file in datapath directory:")
          
            if os.path.exists(evtxdfspklf):
                print("  + Found: " + evtxdfspklf)
            else:
                print("  + Not found.")

        else:
            evtxdfspklf = ""
            extract_records = True

    if evtxdfs_usepickle  and evtxdfspklf != "":
        if os.path.exists(evtxdfspklf):
            # Read from pickle
            print("- Reading raw evtx DFs dictionary (evtxdfs) from pickle file...")
            evtxdfs = pickle.load(open(evtxdfspklf,"rb"))
            print("- Done.")
            extract_records = False
        else:
            if pklinnb :
                print("- WARNING: pickle file not found but notebook cell references it.")
                print("-          Remove from the notebook the cell that contains the "+hostname+"_evtxdfspklf variable definition")
            extract_records = True
    else:
        extract_records = True

    if extract_records :
        # plevtxdf = pldfs['windows_evtx_record']
        print("- Extracting records...")
        evtxdfs = plaso_extract_evtx(pldfs, save_xml=save_xml, xml_filename=xml_filename,files2process=files2process)

        if len(evtxdfs) == 0:
            print("ERROR: No records read.")
            return
        
        # Save to pickle
        if evtxdfs_usepickle :
            if datapath == "":
                print("WARNING: datapath argument not provided. Cannot save pickle file.")
            else:
                print("- Saving evtx DFs dictionary (dfs) as pickle file:")
                print("   "+evtxdfspklf)

                if os.path.exists(datapath):
                    # Save to pickle - - - - - - - - - - - - - - - - - - - - - - - - - -
                    pickle.dump(evtxdfs, open(evtxdfspklf, "wb" ))
                    print("- Done.")

                if savepathtonb :
                    print("- Saving path in jupyter notebook cell...")
                    cell_a = "# Automatically created - DO NOT EDIT OR REMOVE unless you want to change the file to read (in that case, remove this cell)\n"
                    cell_b = "# "+str(datetime.datetime.now())+"\n"
                    cell_c = "# "+hostname+'_evtxdfspklf_f2read = "'+evtxdfspklf+'"'
                    cell   = cell_a + cell_b + cell_c
                    get_ipython().set_next_input(cell)

                    print("  + pickle File path saved to next notebook cell")
                    print("")

    return evtxdfs

# windows_registry_key_value ==================================================
def plaso_extract_win_reg_kv(plobj):
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Input can be either a full plaso DFs dict or a DF of plaso evtx

    d4datatype = "winreg_kv"
    pldatatype = "windows:registry:key_value"

    print("- Extracting plaso "+pldatatype+" events (d4: "+d4datatype+")")
    dfs={}

    if isinstance(plobj, dict):
        pldf = plobj['windows_registry_key_value']
    elif isinstance(plobj, pd.DataFrame):
        pldf = plobj
    else:
        print("ERROR: Unsupported Input Data")
        return

    df = pldf.copy().query('data_type == @pldatatype', engine="python")

    df['D4_DataType_'] = d4datatype

    # Explode values entries into multiple rows (one per value) ---------------

    # Harmonization -----------------------------------------------------------
    print("- Harmonizing")
    df = harmonize_plaso_artifact_common(df)
    df = df.drop(columns=['timestamp', 'date_time'])

    df = df.rename(columns={'Timestamp_': 'KeyLastWriteTimestamp_'})

    df['KeyLastWriteDate_'] = df['KeyLastWriteTimestamp_'].dt.date
    df['KeyLastWriteTime_'] = df['KeyLastWriteTimestamp_'].dt.ceil('s').dt.time

    df['KeyPath_']      = df['key_path']
    df['KeyPath-Hash_'] = df['key_path'].str.lower().apply(hash)
    df['Hive_']         = df['key_path'].str.replace('^(HKEY_LOCAL_MACHINE[^\\\\]*\\\\[^\\\\]*)\\\\.*','\\1')
    df['KeyRelPath_']   = df['key_path'].str.replace('^[^\\\\]*\\\\[^\\\\]*\\\\','')
    df['Values_']       = df['values'].str.replace('^[^\\\\]*\\\\[^\\\\]*\\\\','')
    
    df = df.drop(columns=['key_path', 'values'])

    # Return ------------------------------------------------------------------
    dfs[d4datatype] = df

    print("- Done")

    return dfs

# services - windows_registry_service =========================================
def plaso_extract_win_reg_service(plobj):
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Input can be either a full plaso DFs dict or a DF of plaso evtx

    d4datatype = "winservices"
    pldatatype = "windows:registry:service"

    print("- Extracting plaso "+pldatatype+" events (d4: "+d4datatype+")")
    dfs={}

    if isinstance(plobj, dict):
        pldf = plobj[pldatatype]
    elif isinstance(plobj, pd.DataFrame):
        pldf = plobj
    else:
        print("ERROR: Unsupported Input Data")
        return

    df = pldf.copy().query('data_type == @pldatatype', engine="python")

    df['D4_DataType_'] = d4datatype


    # Harmonization -----------------------------------------------------------
    print("- Harmonizing")
    df = harmonize_plaso_artifact_common(df)
    df = df.drop(columns=['timestamp', 'date_time'])

    #df = df.rename(columns={'Timestamp_': 'KeyLastWriteTimestamp_'})

    df['Date_'] = df['Timestamp_'].dt.date
    df['Time_'] = df['Timestamp_'].dt.ceil('s').dt.time

    df['Name_']           = df['name']
    df['ServiceType_']    = df['service_type']
    df['StartType_']      = df['start_type']
    df['ServiceDLL_']     = df['service_dll']
    df['ObjectName_']     = df['object_name']
    df['ImagePath_']      = df['image_path']
    df['ErrorControl_']   = df['error_control']

    # Common Fields
    df['Group']              = 'd4_niy'
    df['Tag']                = 'd4_niy'
    df['DependOnGroup']      = 'd4_niy'
    df['DependOnService']    = 'd4_niy'
    df['Description']        = 'd4_niy'
    df['DisplayName']        = 'd4_niy'
    df['DriverDisplayName']  = 'd4_niy'
    df['FailureActions']     = 'd4_niy'
    # Specific Fields
    
    df = df.drop(columns=['name', 'service_type', 'start_type', 'service_dll', 'object_name', 'image_path', 'error_control', 'key_path', 'values'])

    # Return ------------------------------------------------------------------
    dfs[d4datatype] = df

    print("- Done")

    return dfs

# amcache =====================================================================

# amcache - windows_registry_amcache =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-

def plaso_extract_win_reg_amcache(plobj):
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    # Input can be either a full plaso DFs dict or a DF of plaso evtx

    d4datatype = "amcache"
    pldatatype = "windows:registry:amcache"

    print("- Extracting plaso "+pldatatype+" events (d4: "+d4datatype+")")
    dfs={}

    if isinstance(plobj, dict):
        pldf = plobj[pldatatype]
    elif isinstance(plobj, pd.DataFrame):
        pldf = plobj
    else:
        print("ERROR: Unsupported Input Data")
        return

    df = pldf.copy().query('data_type == @pldatatype', engine="python")

    df['D4_DataType_'] = d4datatype


    # Harmonization -----------------------------------------------------------
    print("- Harmonizing")
    df = harmonize_plaso_artifact_common(df)
    df = df.drop(columns=['timestamp', 'date_time'])

    #df = df.rename(columns={'Timestamp_': 'KeyLastWriteTimestamp_'})

    df['Date_'] = df['Timestamp_'].dt.date
    df['Time_'] = df['Timestamp_'].dt.ceil('s').dt.time

    df['FileReference_']           = df['file_reference']
    df['FileSize_']                = df['file_size']
    df['FilePath_']                = df['full_path']
    df['LanguageCode_']            = df['language_code']
    df['ProgramIdentifier_']       = df['program_identifier']
    df['CompanyName_']             = df['company_name']
    df['FileDescription_']         = df['file_description']
    df['FileVersion_']             = df['file_version']
    df['ProductName_']             = df['product_name']
    df['SHA1_']                    = df['sha1']

    df = df.drop(columns=['file_reference', 'file_size', 'full_path', 'language_code', 'program_identifier', 'company_name', 'file_description', 'file_version', 'product_name', 'sha1'])

    # Return ------------------------------------------------------------------
    dfs[d4datatype] = df

    print("- Done")

    return dfs

# amcache - windows_registry_amcache_programs =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

# ANALYSIS ####################################################################

# simple ======================================================================
# def simple_func(df,key_path=False,message=False,pathspec=True,ret=False):
def simple_func(df, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")


    if d4.debug >= 4:
        print("DEBUG: [pl] [simple_func()]")

    # Variables ----------------------------------------------------------------
    hiddencolscommon =  ['__container_type__', '__type__', 'data_type', 'inode', 'parser', 'pevtnum', 'message', 'sha256_hash', 'pathspec']
    hiddencolsdf = pd.DataFrame([])

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 1000

    # Arg. Parsing -------------------------------------------------------------
    collapse_constant_cols = kwargs.get('collapse_constant_cols', True)
    hide_cols              = kwargs.get('hide_cols',              True)
    wrap_cols              = kwargs.get('wrap_cols',              True)
    beautify_cols          = kwargs.get('beautify_cols',          False)

    # Show whatever selected on the screen or don't show anything   
    out                    = kwargs.get('out',                    True)
    # Show the resulting DF on the screen
    out_df                 = kwargs.get('out_df',                 True)
    # Return the resulting DataFrame
    ret                    = kwargs.get('ret',                    False)
    # Return the resulting DataFrame as shown in the screen
    # (without hidden cols, collapsed cols, etc.; not in full)
    ret_out                = kwargs.get('ret_out',                False)

    # Check if out is set (will be needed later)
    outargset = 'out' in kwargs.keys()

    # ret_out implies ret
    if 'ret_out' in kwargs.keys() and ret_out:
        ret = True

    # ret implies no out, unless out is set specifically
    if ( ('ret' in kwargs.keys() or 'ret_out' in kwargs.keys()) or 'ret_out' in kwargs.keys()) and not outargset and ret:
        # If ret is set to True by the user we will not provide stdout output,
        # unless the user has specifically set out = True
        out = False

    # Health Check -------------------------------------------------------------
    dfnrows = len(df)

    if dfnrows == 0:
        print("ERROR: Empty DataFrame.")
        return

    # Artifact / Tool specific processing --------------------------------------
    # df.index = df.index.ceil(freq='s')  

    hiddencols = hiddencolscommon
    # pathspec simplification
    if 'pathspec' in df.columns:
        # PREFETCH - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        if   df['parser'].iloc[0] in "prefetch":
            pathspec_simple = df['pathspec'].apply(lambda x: str(x['location']))
            df['pathspec_simple_'] = pathspec_simple
            hiddencols.append('mapped_files')
        # OTHER  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
        elif df['parser'].iloc[0] not in "filestat":
            # Keep a simplified pathspec for certain artifacts
            if 'parent' in df['pathspec'].iloc[0].keys():
                pathspec_simple = df['pathspec'].apply(lambda x: "["+str(x['parent']['location'])+"]"+str(x['location']) if 'location' in x['parent'].keys() else "-").str.replace("/","",1)
                df['pathspec_simple_'] = pathspec_simple

    # Var. Init. ---------------------------------------------------------------
    dfb = df

    # Hidden Columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if dfb['D4_Plugin_'].head(1).values != "windows_registry_key_value":
         hiddencols.append('key_path')

    if hide_cols :
        for col in  hiddencols:
           if col in dfb.columns:
               hiddencolsdf = pd.concat([hiddencolsdf, pd.DataFrame([col])], ignore_index=True)
               dfb = dfb.drop(columns=[col])

    # Constant Columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if collapse_constant_cols :
        concolsdf, dfb = d4utl.collapse_constant_columns(dfb)

    # Wrap wide columns  - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if wrap_cols :
        for col in dfb.columns:
            if isinstance(df[col].iloc[0], str):
                df[col] = df[col].str.wrap(100)
                beautify_cols = True

    # DISPLAY ==========================================================
    nhiddencols = len(hiddencolsdf)

    if out :
        print("")
        display(Markdown("**Statistics:**\n<br>No. Entries: "+str(dfnrows)))

        # Hidden / Constant Columns  - - - - - - - - - - - - - - - - - - - - - - - -
        if collapse_constant_cols :
            show_constant_cols = True
        else:
            show_constant_cols = False

        if hide_cols  and nhiddencols != 0:
            show_hidden_cols = True
        else:
            show_hidden_cols = False

        if show_constant_cols  and show_hidden_cols :
            d4utl.display_side_by_side([hiddencolsdf, concolsdf], ['HIDDEN COLUMNS', 'CONSTANT COLUMNS'])
        elif show_constant_cols  and show_hidden_cols == False:
            display(Markdown("**Constant Columns**"))
            max_rows = pd.get_option("display.max_rows")
            pd.set_option("display.max_rows", None)
            display(concolsdf)
            pd.set_option("display.max_rows", max_rows)
        elif show_constant_cols == False and show_hidden_cols :
            display(Markdown("**Hidden Columns**"))
            display(hiddencolsdf)

        # Beautify columns - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if len(dfb) >= maxdfbprintlines:
            print('WARNING: Too many lines (>'+str(maxdfbprintlines)+') in DataFrame for beautification. Returning non-beautified output.')
            beautify_cols = False

        if out_df :
            if beautify_cols :
                if dfnrows == 1:
                    display(dfb.T.reset_index().style.set_properties(**{'text-align': 'left', 'white-space': 'pre-wrap' }))
                else:
                    display(dfb.reset_index().style.set_properties(**{'text-align': 'left', 'white-space': 'pre-wrap' }))
            else:
                if dfnrows == 1:
                    display(dfb.T)
                else:
                    display(dfb)

    # RETURN  ==========================================================
    if ret :
        if ret_out :
            return dfb
        else:
            return df
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # plaso original 
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #key_path = kwargs.get('key_path',False)
    #message  = kwargs.get('message',False)
    #pathspec = kwargs.get('pathspec',True)


def get_source_options():
    return ['wrap_cols', 'beautify_cols']

# analysis ====================================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    return analysis_func(*args, **kwargs)

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """
    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-list")
        return

    def d4list(objtype):
        print("Available plaso analysis types:")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.search("^dict-pandas_dataframe-plaso", objtype):
            print("- plaso_categories:  No.events & first/last event per plaso category (Input: pldfs)")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.search("^dict-pandas_dataframe-plaso", objtype):
            print("- plaso_overview: Overview (Input: pldfs)")

    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # ANALYSIS FUNCTIONS ======================================================

    # pldfs ------------------------------------------------------------------
    if re.search("^dict-pandas_dataframe-plaso", objtype):
        if anltype == "plaso_categories":
            return analysis_plaso_categories(*args, **kwargs)
        if anltype == "plaso_overview":
            return analysis_plaso_overview(*args, **kwargs)

    print("INFO: [d4pl] No analysis functions available for this data type ("+objtype+")")

def analysis_plaso_categories(*args, **kwargs):
    """ Analysis that gives plaso categories

        Args: 
        obj:          Input data (typically DF or dict of DFs)
        Returns: 
        pandas.Dataframe with the results of the analysis

    """
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    dfs = args[0]

    objtype = d4com.data_identify(dfs)

    if objtype != "dict-pandas_dataframe-plaso-raw":
        print("ERROR: Invalid object for function: "+objtype)
        print("       Input object should be:      dict-pandas_dataframe-plaso")
        return

    outdf = pd.DataFrame([],columns=['Category','NEvts','TSMin','TSMax'])
    row = pd.Series()

    keys = list(dfs.keys())
    keys.sort()

    for key in keys:
        if len(dfs[key]) != 0:
            row['TSMin'] = dfs[key]['Timestamp_'].min()
            row['TSMax'] = dfs[key]['Timestamp_'].max()
        else:
            row['TSMin'] = ""
            row['TSMax'] = ""

        row['Category'] = key
        row['NEvts'] = len(dfs[key])

        outdf = outdf.append(row,ignore_index=True)

    return outdf

def analysis_plaso_overview(*args, **kwargs):
    """ Analysis that gives general device information

        Args: 
        obj:          Input data (typically DF or dict of DFs)
        Returns: 
        pandas.Dataframe with the results of the analysis

    """    
    if d4.debug >= 3:
        print("DEBUG: [DBG"+str(d4.debug)+"] ["+str(os.path.basename(__file__))+"] ["+str(inspect.currentframe().f_code.co_name)+"()]")

    dfs = args[0]
    
    objtype = d4com.data_identify(dfs)
    
    if objtype != "dict-pandas_dataframe-plaso-raw":
        print("ERROR: Invalid object for function: "+objtype)
        print("       Input object should be:      dict-pandas_dataframe-plaso")
        return
    
    df = pd.DataFrame()
    overview_df = pd.DataFrame()
    users_df = pd.DataFrame()
    
    for key in dfs:
        if key == 'windows_registry_installation':
            df = dfs[key]
            overview_df['Product Name'] = df['product_name']
            overview_df['Installation Date'] = df['Timestamp_'].iloc[1]
            
        if key == 'windows_registry_timezone':
            df = dfs[key]
            overview_df['Timezone'] = df['configuration']
        
        if key == 'windows_registry_sam_users':
            df = dfs[key]
            users_df['User Name'] = df['username']
            users_df['First Connection'] = df['Timestamp_']
            
    outdf = overview_df.iloc[0].to_frame()
    
    outdf['System Information'] = outdf[0]
    
    outdf = outdf.drop(columns=[0,])
    
    display(outdf)
    
    return users_df
    
    

# DATAFRAME ACCESSORS =========================================================

@pd.api.extensions.register_dataframe_accessor("d4pl")
class Ds4n6PlAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df=self._obj

        return simple_func(df, **kwargs)

    def regfindraw(self,str,full=False):
        df=self._obj
        if full :
            return df.fillna("null_value").query('values.str.contains(@str)',engine="python")
        else:
            return df.fillna("null_value").query('values.str.contains(@str)',engine="python").d4plaso.simple().drop(columns=['values','timestamp_desc'])

    def regfind(self,searchstr):
        df=pd.DataFrame(self.regfindraw(searchstr,full=True)['message'].apply(lambda x: re.sub(r"([A-Za-z][A-Za-z]*:) (\[[A-Z_][A-Z_]*\])","-=- \\1 | \\2 | ",x)))
        html0=df.to_html()
        html1=re.sub("</td>","</td></tr></table></td>",re.sub("<td>","<td><table><tr><td>",html0))
        html2=re.sub("] -=- ","]</td><td></td><td></td></tr><tr><td>",html1)
        html3=re.sub(" -=- ","</td></tr><tr><td>",html2)
        html4=re.sub(r"\|","</td><td>",html3)
        html=html4
        display(HTML(html))

@pd.api.extensions.register_dataframe_accessor("d4_plaso")
class Ds4n6PlasoAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df=self._obj

        return simple_func(df, **kwargs)

    def regfindraw(self, str, full=False):

        df = self._obj

        if full :
            return df.fillna("null_value").query('values.str.contains(@str)',engine="python")
        else:
            return df.fillna("null_value").query('values.str.contains(@str)',engine="python").d4plaso.simple().drop(columns=['values','timestamp_desc'])

    def regfind(self, searchstr):
        df=pd.DataFrame(self.regfindraw(searchstr,full=True)['message'].apply(lambda x: re.sub(r"([A-Za-z][A-Za-z]*:) (\[[A-Z_][A-Z_]*\])","-=- \\1 | \\2 | ",x)))
        html0 = df.to_html()
        html1 = re.sub("</td>","</td></tr></table></td>",re.sub("<td>","<td><table><tr><td>",html0))
        html2 = re.sub("] -=- ","]</td><td></td><td></td></tr><tr><td>",html1)
        html3 = re.sub(" -=- ","</td></tr><tr><td>",html2)
        html4 = re.sub(r"\|","</td><td>",html3)

        html = html4

        display(HTML(html))




# DS4N6
#
# Description: Library of functions to apply Data Science to forensics artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4mcrb

###############################################################################
# IMPORTS
###############################################################################

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import inspect
import pickle

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui
import ds4n6_lib.utils  as d4utl
# bug: unix no exist error, replace by unx.
import ds4n6_lib.unx   as d4unx

###############################################################################
# FUNCTIONS
###############################################################################

# Hidden columns in simple() funcion
hiddencols = [ 'MTStampEpoch_', 'MTStamp_', 'ATStampEpoch_', 'ATStamp_', 'CTStampEpoch_', 'CTStamp_', 'Meta_', 'FileStem_', 'ParentPath_', 'ParentName_', 'PathSeparator_', 'FilePath-Hash_', 'FileName-Hash_', 'FileStem-Hash_', 'ParentPath-Hash_', 'ParentName-Hash_'] 

###############################################################################
# FUNCTIONS
###############################################################################

# FILE READING FUNCTIONS ######################################################

def read_data(evdl, **kwargs):
    """ Read data from files or a folder

        Args: 
            evdl (str): path to file/folder source
            kwargs: read options
        Returns: 
            pandas.Dataframe or dictionary of pandas.DataFrame
    """
    if d4.debug >= 3:
        print("DEBUG: [macrobber-read_data()]")

    header_names = ['MD5', 'path', 'inode', 'mode_as_string', 'UID', 'GID', 'size', 'atime', 'mtime', 'ctime', 'block_size'] 

    kwargs['header_names'] = header_names

    return d4com.read_data_common(evdl, **kwargs)

# HARMONIZATION FUNCTIONS #####################################################

def harmonize(df, **kwargs):
    """ Convert DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    data_os           = kwargs.get('data_os',           None)
    generate_hashes   = kwargs.get('generate_hashes',   True)
    path_prefix       = kwargs.get('path_prefix',       None)

    # Specific Harmonization Pre-Processing ===================================
    def remove_prefix(df, prefixregex):
        if 'FilePath_' in df.columns:
            df['FilePath_'] = df['FilePath_'].str.replace(prefixregex,'')
        return df

    # Harmonize to File_List_HAM

    # PathSeparator is tool-dependent, not only OS-dependent
    pathsep = '/'

    df['MTStampEpoch_']  = df['mtime']
    df['MTStamp_']       = pd.to_datetime(df['mtime'], errors = 'coerce',  unit='s')
    df['MTStampDate_']   = df['MTStamp_'].dt.date
    df['MTStampTime_']   = df['MTStamp_'].dt.ceil(freq='s').dt.time
    df['MTStampDoW_']    = df['MTStamp_'].dt.day_name()
    df['ATStampEpoch_']  = df['atime']
    df['ATStamp_']       = pd.to_datetime(df['atime'], errors = 'coerce',  unit='s')
    df['ATStampDate_']   = df['ATStamp_'].dt.date
    df['ATStampTime_']   = df['ATStamp_'].dt.ceil(freq='s').dt.time
    df['ATStampDoW_']    = df['ATStamp_'].dt.day_name()
    df['CTStampEpoch_']  = df['ctime']
    df['CTStamp_']       = pd.to_datetime(df['ctime'], errors = 'coerce',  unit='s')
    df['CTStampDate_']   = df['CTStamp_'].dt.date
    df['CTStampTime_']   = df['CTStamp_'].dt.ceil(freq='s').dt.time
    df['CTStampDoW_']    = df['CTStamp_'].dt.day_name()
   #df['BTStampEpoch_']  = df['btime']
   #df['BTStamp_']       = pd.to_datetime(df['btime'], errors = 'coerce',  unit='s')
   #df['BTStampDate_']   = df['BTStamp_'].dt.date
   #df['BTStampTime_']   = df['BTStamp_'].dt.ceil(freq='s').dt.time
   #df['BTStampDoW_']    = df['BTStamp_'].dt.day_name()
    df['Size_']          = df['size'].astype('int64')
   #df['Mode_']          = None
    if not data_os == "windows":
        df['UID_']       = df['UID']
    if not data_os == "windows":
        df['GID_']       = df['GID']
    df['Meta_']          = df['inode']
   #df['File_Name']      = None
    df['Type_']          = df['mode_as_string'].str.extract('^(.)')
   #df['PrevType_']      = None
    if not data_os == "windows":
        df['Permissions_'] = df['mode_as_string'].str.replace('^.','').str.replace(r'\ .*$','')
   #df['Deleted_']       = None
   #df['Reallocated_']   = None
   #df['Hostname_']      = None
    if not df['MD5'].iloc[0] == 0:
        df['MD5_Hash_']  = df['MD5']
   #df['SHA256_Hash_']   = None
   #df['DriveLetter_']   = None
   #df['VSS_']           = None
   #df['EVOName_' ]      = None
   #df['EvidenceName_']  = None
   #df['Partition_']     = None
   #df['Tag_']           = None
    df['FilePath_']      = df['path']
    if path_prefix is not None:
        df = remove_prefix(df, path_prefix)
    df['FileName_']      = df['FilePath_'].str.replace('.*'+pathsep,'')
    df['FileStem_']      = df['FileName_'].str.replace(r'\.[^\.]*$','')
    df['FileExtension_'] = df['FileName_'].str.replace(r'^[^\.]*$', '').str.replace(r'.*\.','').str.lower()
    df['ParentPath_']    = df['FilePath_'].str.replace('(.*)'+pathsep+'.*','\\1')
    df['ParentName_']    = df['ParentPath_'].str.replace('.*'+pathsep,'')
    df['PathSeparator_'] = pathsep
   #df['FSType_']        = None
   #df['TSNTFSAttr_']    = None


    # Path-Hash Fields  - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    if generate_hashes:
        df['FilePath-Hash_']   = df['FilePath_'].str.lower().apply(hash)
        df['FileName-Hash_']   = df['FileName_'].str.lower().apply(hash)
        df['FileStem-Hash_']   = df['FileStem_'].str.lower().apply(hash)
        df['ParentPath-Hash_'] = df['ParentPath_'].str.lower().apply(hash)
        df['ParentName-Hash_'] = df['ParentName_'].str.lower().apply(hash)

    # Generic Harmonization ===================================================
    df = d4com.harmonize_common(df, datatype='flist', **kwargs)

    # Specific Harmonization Post-Processing ==================================

    # return ==================================================================

    return df

# ANALYSIS FUNCTIONS ##########################################################

# simple ======================================================================
def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """

    if d4.debug >= 4:
        print("DEBUG: [mcrb] [simple_func()]")

    # Artifact-specific argument parsing =======================================

    # Variables ================================================================
    dfout = df

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Pre-Processing ==========================================================

    # Call to simple_common ===================================================
    dfout = d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)

    # Post-Processing =========================================================

    # Return ==================================================================
    return dfout

# analysis ====================================================================
def analysis(obj, *args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    return analysis_func(obj, *args, **kwargs)

def analysis_func(obj, *args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """

    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        anlav = False
        print("Available macrobber analysis types:")
        if objtype == None or objtype == "str-help" or objtype == "str-list" or  re.search("^dict-pandas_dataframe-macrobber", objtype):
            anlav = True
            print("- macrobber_files:  No.events macrobber file (Input: macrobberdfs)")

        if anlav == False:
            print('- No analysis modules available for this object ('+objtype+').')

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # ANALYSIS FUNCTIONS ======================================================

    # mcrbdfs ------------------------------------------------------------------
    # if   re.search("^dict-pandas_dataframe-macrobber", objtype):
    #     if anltype == "macrobber_files":
    #         return analysis_macrobber_files(*args, **kwargs)

    print("INFO: [d4mcrb] No analysis functions available for this data type ("+objtype+")")

# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4mcrb")
class Ds4n6McrbAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)


@pd.api.extensions.register_dataframe_accessor("d4_macrobber")
class Ds4n6MacRobberAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)




#
# Description: library of functions to appy Data Science in several forensics
#              artifacts
#

###############################################################################
# INFO
###############################################################################
# Recommended "import as": d4mctm

###############################################################################
# IMPORTS
###############################################################################

# DEV  IMPORTS ----------------------------------------------------------------

# python IMPORTS --------------------------------------------------------------
import os
import glob
import re
import time
import inspect
import xmltodict
import json
import pickle
from tqdm import tqdm
import xml.etree.ElementTree as et

# DS IMPORTS ------------------------------------------------------------------
import numpy  as np
import pandas as pd
import matplotlib.pyplot as plt
from IPython.display import display, Markdown, HTML

from sklearn.model_selection import train_test_split
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.layers import Input, Dense

# DS4N6 IMPORTS ---------------------------------------------------------------
import ds4n6_lib.d4     as d4
import ds4n6_lib.common as d4com
import ds4n6_lib.gui    as d4gui
import ds4n6_lib.utils  as d4utl

###############################################################################
# IDEAS
###############################################################################
# is_deleted()
# is_file()
# is_dir() / is_folder() - level
# ext() # filter by Extension
# nofn  # exclude $FILE_NAME entries

###############################################################################
# FUNCTIONS
###############################################################################

# FILE READING FUNCTIONS ######################################################

# FILE READING FUNCTIONS ######################################################

def read_data(evdl, **kwargs):
    if d4.debug >= 3:
        print("DEBUG: [mctm] read_data")

    return d4com.read_data_common(evdl, **kwargs)

# HARMONIZATION FUNCTIONS #####################################################

def harmonize(df, **kwargs):
    """ Convert DF in HAM format

        Args: 
            df (pandas.DataFrame): DF to harmonize
            kwargs(dict): harmonize options
        Returns: 
            pandas.DataFrame in HAM Format
    """
    objtype = d4com.data_identify(df)

    if objtype == "pandas_dataframe-mactime-raw":
        # Specific Harmonization Pre-Processing -----------------------------------
        df = df.rename(columns={"Type": "MACB"})

        df['Type_']        = df['Mode'].str.extract('^(.)')
        df['PrevType_']    = df['Mode'].str.extract('^..(.)')
        df['Permissions_'] = df['Mode'].str.extract('^...(.........)')

        # Deleted / Reallocated
        df['Deleted_']     = df['File Name'].str.contains(r'\ \(deleted\)$|\ \(deleted-reallocated\)$')
        df['Reallocated_'] = df['File Name'].str.contains(r'\ \(deleted-reallocated\)$')

        # [FT] Tag -> Tag_ | DriveLetter_ | VSS_ | EVOName_ | EvidenceName_ | Partition_ | FilePath_
        # FT
        if re.search(r'^[A-Z]\[vss[0-9][0-9]\]{.*}:', df['File Name'].iloc[0]):
            fncolsdf  = df['File Name'].str.split(":", 1, expand=True).rename(columns={0: "Tag_", 1: "FilePath_"})
            fncolsdf['FilePath-Hash_'] = fncolsdf['FilePath_'].str.lower().apply(hash)
            fncolsdf['FSType_']   = '-'
            df['Hostname_']    = '-'
            df['SHA256_Hash_'] = '-'

            fncols2df = fncolsdf['Tag_'].str.extract(r'([A-Z])\[vss(.*)\]{(.*)}', expand=True).rename(columns={0: "DriveLetter_", 1: "VSS_", 2: "EVOName_"})
            fncols2df['VSS_'] = fncols2df['VSS_'].astype(int)

            fncols3df = fncols2df['EVOName_'].str.extract('(.*)-ft-p(.*)', expand=True).rename(columns={0: "EvidenceName_", 1: "Partition_"})
            fncols3df['Partition_'] = fncols3df['Partition_'].astype(int)

            df = pd.concat([df, fncols2df, fncols3df, fncolsdf], axis=1)

        else:
            fncolsdf  = df['File Name'].str.split(":", 1, expand=True).rename(columns={0: "Tag_", 1: "FilePath_"})
            df = pd.concat([df, fncolsdf], axis=1)
            df['Hostname_']     = '-'
            df['EVOName_']      = '-'
            df['EvidenceName_'] = '-'
            df['Partition_']    = '-'
            df['FSType_']       = '-'
            df['DriveLetter_']  = '-'
            df['VSS_']          = '-'
            df['TSNTFSAttr_']   = '-'
            df['SHA256_Hash_']  = '-'

        # Deal with "($FILE_NAME)" string
        tsntfsattrmap = {True: 'FILE_NAME', False: 'STD_INFO'}
        df['TSNTFSAttr_']  = df['FilePath_'].str.contains(r'\ \(\$FILE_NAME\)$').map(tsntfsattrmap)
        df['FilePath_']    = df['FilePath_'].str.replace(r'\ \(\$FILE_NAME\)$','')

        df['FilePath_'] = df['FilePath_'].str.replace(r'\ \(deleted\)$|\ \(deleted-reallocated\)$','')
        
        # Generic Harmonization ---------------------------------------------------
        df = d4com.harmonize_common(df, **kwargs)

        # Specific Harmonization Post-Processing ----------------------------------

        return df

# CORE FUNCTIONS (simple, analysis, etc.) #####################################

# simple ======================================================================

def simple_func(df, *args, **kwargs):
    """ Reformat the input df so the data is presented to the analyst in the
        friendliest possible way

    Parameters:
    df  (pd.dataframe):  Input data 
    
    Returns:
    pd.DataFrame: Optionally it will return the filtered dataframe, 
                  only if ret=True is set, constant & hidden columns included
                  If ret_out=True is set, then the output just as it is shown
                  (without constant/hidden columns) will be return
    """

    if d4.debug >= 3:
        print("DEBUG: [mctm] [simple_func()]")

    windows = kwargs.get('windows', True) 

    # Variables ----------------------------------------------------------------
    hiddencols =  ['File_Name', 'FilePath-Hash_', 'SHA256_Hash_']

    if windows :
        nonwincols = ['UID', 'GID', 'Mode', 'Permissions_']
        hiddencols = hiddencols + nonwincols

    # Maximum number of lines in DF for beautification
    maxdfbprintlines = 20

    # Call to simple_common ----------------------------------------------------
    return d4com.simple_common(df, *args, **kwargs, hiddencols=hiddencols, maxdfbprintlines=maxdfbprintlines)


# DATAFRAME ACCESSOR ##########################################################

@pd.api.extensions.register_dataframe_accessor("d4mctm")
class Ds4n6MctmAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)

@pd.api.extensions.register_dataframe_accessor("d4_mactime")
class Ds4n6MactimeAccessor:
    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def simple(self, *args, **kwargs):
        """ Redirects execution to simple_func()
        """
        df = self._obj
        return simple_func(df, *args, **kwargs)

# ANALYSIS ####################################################################

# analysis() function =========================================================
def analysis(*args, **kwargs):
    """ Redirects execution to analysis_func()
    """
    return analysis_func(*args, **kwargs)

def analysis_func(*args, **kwargs):
    """ Umbrella function that redirects to different types of analysis 
        available on the input data

    Parameters:
    obj:          Input data (typically DF or dict of DFs)
    
    Returns:
    pd.DataFrame: Refer to each specific analysis function
    """

    def syntax():
        print('Syntax: analysis(obj, "analysis_type")\n')
        d4list("str-help")
        return

    def d4list(objtype):

        # Analysis Modules Available for this objective
        # anlav = False
        print("Available fstl analysis types:")
        print("- No analysis functions defined yet.")
        return

        # TEMPLATE
        #if objtype == "str-help" or objtype == "str-list" or  re.search("^pandas_dataframe-fstl-mactime-standard", objtype):
        #    anlav = True
        #    print("- XXXXXXXXXX:  XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX (Input: fstldf)")

        # if anlav == False:
        #     print('- No analysis modules available for this object ('+objtype+').')

    nargs = len(args)

    if nargs == 0:
        syntax()
        return

    obj = args[0]

    objtype = d4com.data_identify(obj)

    if isinstance(obj, str):
        if obj == "list":
            d4list(objtype)
            return
        if obj == "help":
            syntax()
            return

    if nargs == 1:
        syntax()
        return

    anltype = args[1]

    if not isinstance(anltype, str):
        syntax()
        return

    if anltype == "help":
        syntax()
        return
    elif anltype == "list":
        d4list(objtype)
        return

    # TEMPLATE
    # If object is a dict of dfs
    #elif re.search("^pandas_dataframe-evtx_file_df", objtype):
    #    if anltype == "XXXXXXXXXXX":
    #        return XXXXXXXXXXXXXXXXXXXXX(*args, **kwargs)
    #else:
    #    print("ERROR: [fstl] Unsupported input data.")
    #    return




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ds4n6_lib",
    version="0.7.1",
    author="Jess Garcia",
    author_email="ds4n6@one-esecurity.com",
    description="Bringing Data Science & Artificial Intelligence to the fingertips of the average Forensicator, and promote advances in the field",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ds4n6/ds4n6_lib",
    project_urls={
      "Bug Tracker" : "https://github.com/ds4n6/ds4n6_lib/issues",
      "Website"     : "http://www.ds4n6.io/"
    },
    keywords = ['dfir', 'datascience', 'forensics'],
    install_requires=[
          'requests',
          'numpy',
          'pandas',
          'Evtx',
          'python-evtx',
          'ipyaggrid',
          'IPython',
          'ipywidgets',
          'keras',
          'matplotlib',
          'nbformat',
          'numpy',
          'pandas',
          'pyparsing',
          'qgrid',
          'ruamel.yaml',
          'sklearn',
          'tensorflow',
          'tqdm',
          'traitlets',
          'xmltodict',
      ],
    classifiers=[
      "Development Status :: 3 - Alpha",
      "Intended Audience :: Developers",
      "Intended Audience :: Information Technology",
      "Framework :: Jupyter",
      "Topic :: Security",
      "Topic :: Scientific/Engineering :: Artificial Intelligence",
      "Topic :: Software Development :: Libraries :: Python Modules",
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
      "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)
